# Atomic and Molecular Modeling Language

The Atomic and Molecular Modelling Language (AMML) is designed as a modular extension of the [Sientific Computing Language (SCL)](scl.md).

The language provides builtin types, expressions and functions specific for atomistic modeling and simulation.

## Structure

An atomic structure can be defined using the `Structure` literal. It has the following syntax 

```
Structure [optional name] from <Table>
```

The `<Table>` parameter is a [`Table` literal](scl.md#table) with specific constraints listed in the following.

name    | description   | data struct | type           | shape | units / dimension | 
--------|---------------|-------------|----------------|-------|-------------------|
atoms   | individual atoms structures | Series | Table | columns: symbols, x, y, z, px, py, pz, tags, masses | None
cell    | unit cell   | Series | FloatArray | (3,3)  | [length] |
pbc     | periodic boundary conditions  | Series | BoolArray |(3) | None |

Every tuple (row, record) in the top-level `Table` corresponds to an individual atomic, or molecular, structure.
The atoms column is a special case: every element of the atoms column, is also a Table whose own columns hold the properties of chemical species such as: atomic symbols, positions, masses, momenta, etc. So the Table parameter of the Structure literal contains the atoms Table nested within. The columns of the atoms sub-table have the following constraints:

name    | description   | data struct | type           | shape | units / dimension | 
--------|---------------|-------------|----------------|-------|-------------------|
symbols | the chemical symbols | Series | String | scalar | None |
x       | the cartesian x position | Series | Float | scalar | [length] |
y       | the cartesian y position | Series | Float | scalar | [length] |
z       | the cartesian z position | Series | Float | scalar | [length] |
px      | cartesian momentum along x | Series | Float | scalar |  [length*mass/time] |
py      | cartesian momentum along y | Series | Float | scalar |  [length*mass/time] |
pz      | cartesian momentum along z | Series | Float | scalar |  [length*mass/time] |
masses  | the atomic mass | Series | Float | scalar | [mass] |
tags    | tag      | Series | Integer | scalar | [dimentionless] |


Example:

```
water = Structure H2O from (
         (atoms: ((symbols: 'O', 'H', 'H'),
                  (x: 0., 0., 0.) [nm],
                  (y: 0., 0.763239, -0.763239) [angstrom],
                  (z: 0.119262, -0.477047, -0.477047) [angstrom],
                  (tags: 1, 0, 0),
                  (masses: 16., 1., 1.) [amu]
                 )
         ),
         (cell: [[5., 0., 0.], [0., 5., 0.], [0., 0., 5.]] [bohr]),
         (pbc: [false, false, false])
       )
```

An atomic structure can be loaded from a file, e.g.:

```
water = Structure from file 'h2o.xyz'
```

Note that, when loading from file, there would be no `water.name` parameter.

The parameters contained in a `Structure` parameter can be accessed using the common [subscripting](scl.md#subscripting) or [slice](scl.md#slice) syntax:

``` 
print(water.name) # name of structure
print(water.atoms) # the series with atoms tables
print(water.atoms[0]) # atoms table of the first structure
print(water.atoms[0:1]) # series with atoms inluding only the first structure
print(water.pbc) # series of boolean arrays
print(water.cell) # series of float arrays
print(water.cell[0]) # the cell of the first structure as float array
print(water[0]) # a tuple (atoms, cell, pbc) of the first structure
print(water[0:1]) # a table containing the first structure
```

## Calculator

A `Calculator` parameter describes operations to be performed on a [Structure](#structure) parameter typically employing computational tools external to the AMML, e.g. molecular dynamics or quantum chemistry packages. It is commonly used within the [Property](#property) parameter. The `Calculator` has the following syntax:

```
Calculator <name> [ >|<|==|>=|<= <version>] <Table>
Calculator <name> [ >|<|==|>=|<= <version>] (default)
Calculator <name> [ >|<|==|>=|<= <version>] ()
```

The specification of `name` is mandatory. Currently `name` must be one of `vasp`, `turbomole`, `lj`, `emt`.

The version specification is optional. Currently it is not interpreted.

The actual calculator parameters can be specified as series in the top-level table. The paramater names, type and units depend on the calculator. The table below lists the parameters of the currently supported calculators. If only default parameters have to be used then no `Table` parameter is needed and the syntax is `(default)` or simply `()`.

Example for a VASP calculator with one parameter set:

```
calc = Calculator vasp == 5.4.4 (
        (algo: 'Fast'),
        (ediff: 1e-06) [eV],
        (ediffg: -0.01) [eV/angstrom],
        (encut: 400.0) [eV],
        (ibrion: 2),
        (icharg: 2),
        (isif: 2),
        (ismear: 0),
        (ispin: 2),
        (istart: 0),
        (kpts: [5, 5, 1]),
        (lcharg: false),
        (lreal: 'Auto'),
        (lwave: false),
        (nelm: 250),
        (nsw: 1500),
        (potim: 0.1),
        (prec: 'Normal'),
        (sigma: 0.1) [eV],
        (xc: 'PBE')
)
```

**NOTE:** The AMML language can be extended to support further calculators. This is fairly straight forward because the current implementation of AMML interpreter is based on the [Atomic Simulation Environment (ASE)](https://wiki.fysik.dtu.dk/ase/). Particularly, the AMML `Calculator` parameter can reuse any existing ASE Calculator class. The process of integration of further calculators into AMML is basically a registration of the name in the grammar and of function to check calculator parameter types and units.

Similar to AMML `Structure` also a `Calculator` parameter can be loaded from a file:

```
calc = Calculator from file 'emt-custom.yaml'
```

The parameters stored in the calculator can be accessed using the [subscripting](scl.md#subscripting) and [slice](scl.md#slice) syntax:

```
print(calc.parameters) # top-level table of parameters 
print(new.name) # calculator's name
print(calc.pinning, calc.version) # if not set `null` will be printed
print(calc[0]) # a tuple of the first parameter set
print(new_calc[0:1]) # table with the first parameter set
```

**NOTE:** The calculator parameter names and types are identical with those in the [Atomic Simulation Environment (ASE)](https://wiki.fysik.dtu.dk/ase/ase/calculators/calculators.html). For this information we refer to the ASE documentatoin of the calculators. The physical units (or their dimension) of the calculator parameters are not always fully and systematically documented. While units / dimensionality checks are already performed independently by the AMML interpreter (with error messages suggesting relevant corrections), this information is not yet provided here. For now, the user of AMML should consult the manuals of the original codes.

## Constraint

The `Constraint` parameter describes constraints to be applied to atomic positions. A constraints is pertaining to (but not an integral part of) a specific `Structure` parameter.

```
c1 = FixedAtoms where <fixed>
c2 = FixedLine collinear to <direction> where <fixed>
c3 = FixedPlane normal to <direction> where <fixed>
```

The `<fixed>` parameter is a `Series` of `Boolean` type, 
whose length should be equal to the number of atoms in the structure to which the constraint will be applied. It indicates which atoms are included in the constraint (`true` elemenets) and which are not (`false` elements).

The `<direction>` parameter is an `Array` of `Integer` type of length 3 and represents a vector in cartesian space `[x, y, z]`.

Examples:
```
h2o_plane = FixedPlane normal to [0, 0, 1] where (fix: true, true, true) # allow changes in xy plane
h2_line = FixedLine collinear to [0, 0, 1] where (fix: true, true) # allow changes along z axis
h2_fix0 = FixedAtoms where (fix: true, false) # fix first atom in H2
```

**NOTE:** Extensions of the AMML with further types of contstraints is possible.


## Property

Using the AMML `Property` parameter a set of properties will be computed for a specific AMML `Structure` with a specific *method*. The method can be specified by an AMML `Calculator` and a `task`.

Syntax:
```
prop = Property name1[, name2[...]] 
for structure <struct> with calculator <calc> 
[with constraints (c1[, c2[, ...]] )] [, task: <task_name>]
```

The names of currently intepreted properties are summarized in the table below. The parameters `<struct>`, `<calc>`, and `c1`, ... are names of variables of the corresponding types `Structure`, `Calculator` and `Constraint` respectively. The `task_name` can be currently one of `single point`, `local minimum`, `global minimum`, `transition state`, `normal modes`, `micro-canonical`,`canonical`, `isothermal-isobaric`, `grand-canonical`. Currently, the task name is only used to perform consistency checks.

propery name          | type            | units (of output) | description
----------------------|-----------------|-------------------|------------ 
energy_minimum        | Boolean         | None              | `true` if geometry is minimum `false` otherwise
energy                | Float           | [eV]              | the total potential energy of the system
forces                | FloatArray      | [eV / angstrom]   | the forces for each atom, i.e. the first derivative of energy with a minus sign
dipole                | FloatArray      | [elementary_charge * angstrom] | the total  electrostatic dipole moment of the system
vibrational_energies  | Series(Float)   | [eV]              | the eigen energies ("frequencies") of the harmonic normal modes


After initializing a variable with a `Property` parameter, the individula properties can be accessed using subscripting.

Examples:
```
props = Property energy, forces for structure h2o with calculator calc,
                 task: local minimum

print(props) # print the whole property parameter
print(props[0]) # return the first tuple of the property parameter
print(props.struct) # return the initial Structure parameter
print(props.calc) # return the initial Calculator parameter
print(props.names) # return the property names
print(props.task) # return the property task
print(props.calc[0:1]) # return a slice of the initial calc 
print(props.calc.name) # return the name of the inital calc
print(props.calc.parameters) # return a Table of the calc parameters
print(props.calc.version) # return the calc version
print(props.results) # return a Table with all calculated properties
print(props.output_structure) # return the output Structure
print(props.output_structure[0]) # return the first output Structure
print(props.energy) # return energy as Series for all output structures
print(props.energy[0:1]) # return a slice (Series) with the energy of the first output structure
print(props.energy[0]) # return energy (Float) of the first output structure
print(props.forces) # return forces as series for all output structures
print(props.forces[0]) # return an Array of forces for the first output structure 
```

The `Property` parameter is evaluated in the following way. The input [`Structure`](#structure) and [`Calculator`](#calculator) parameters are subsciptable and contain Series of structure geometries and calculator parameters, respectively. The cartesian product of these two Series is taken to find all combinations and for each combination a calculation is performed. Every calculation is captured as a tuple (a row) in `props.results` Table. The `props.results` Table has as columns the names of all requested properties, as well as the corresponding input geometry, calculator parameters and final (output) geometry.

For example, if two geometries of the same molecule have to be processed by the same calculator with three different parameter sets, then the `results` Table will contain six tuples. The constraints, the task and the properties computed are the same for all struct-calc combinations. The results can be accessed via subscripting the `Property` parameter, as it is shown in the examples above.
