# Supporting tools

To enable the usage of the SCL and AMML, a minimum set of supporting tools are necessary. These tools include a parser, a correctness checker, an interpreter, and a syntax highlighter. All these tools (or subsets of them) can be used via the command-line interface or the Jupyter/iPython kernel.

## Using the command-line interface

The command-line interface is provided by the scripts located in the `scripts` folder.

### Script mode

The script `run_model.py` runs a DSL *program* (also called *model*) in one of three different evaluation modes. The basics on how to write a program in the Scientific Computing Language can be found in [this section](scl.md) and Atomic and Molecular Modeling Language in [this section](amml.md).

In all evaluation modes, the model is first parsed using a grammar. The grammar can be optionally specified by the flag `-g GRAMMAR_PATH, --grammar_path`. The default grammar used for parsing is in the file `grammar/virtmat.tx`. Syntax errors are issued in this phase. The interpreter then starts applying static constraints such as type checks. Static errors (such as type errors) are issued at this stage. Then the evaluation of the model is performed. The evaluation (or execution) mode can be specified using the `-m | --mode` option. These modes are *instant*, *deferred*, and *workflow*. The instant evaluation mode is the default. More details are available in the following subsections. The *model source file* (with extension `.vm` ) must always be specified using the flag `-f MODEL_FILE, --model_file MODEL_FILE` even if it is empty.

#### Instant and deferred modes

In the *instant* mode the evaluation is performed *immediately* and *locally*. In *deferred* mode the evaluation is deferred but still performed *locally*. In *instant* and *deferred* modes the model does not persist after the interpreter has finished, i.e. the script *run_model.py* has returned. Only parameters of `print` statements in the program will be evaluated. For debugging purposes, the flag `-p, --print-all-vars` can be specified with which all variables will be evaluated and their names, values and types will be printed.

#### Workflow mode

In the *workflow* mode the intepreter creates a workflow from the model and stores the workflow in a database. Therefore, the model is *persistent* after the interpreter returns, i.e. the *model instance* continues existing on the database. A model instance can be retrieved via a *universally unique identifier* (UUID) that is generated at the time when the model instance is created. A model instance can only be evaluated and extended but not otherwise modified.

If no additional flags are specified the model is not evaluated after interpretation except for parameters in `print` statements that contain no references to variables. Otherwise `n.c.` will be printed instead of the not yet computed values. Optionally, a model can be evaluated locally after the intepreter finishes by specifying the ` -a | --autorun` flag. In this case all variables will be evaluated but only those in print statements will be printed. Additionally, the `-d, --on-demand` flag restricts the evaluation to only those variables that are referenced in `print` statements.

In order to work in workflow mode a launchpad (a MongoDB database)must be configured. Using the flag `-l LAUNCHPAD_FILE, --launchpad_file LAUNCHPAD_FILE` the launchpad data (such as hostname, database name and credentials) can be optionally specified. Otherwise the script looks up the current working directory and then the folder `$HOME/.fireworks` for a launchpad configuration. More information about configuring a launchpad can be found [here](https://git.scc.kit.edu/virtmat-tools/vre-middleware/-/blob/master/docs/launchpad.md).

By using the `-h, --help` flag a summarized help can be obtained:

```
python scripts/run_model.py --help
usage: run_model.py [-h] -f MODEL_FILE [-g GRAMMAR_PATH] [-m {instant,deferred,workflow}] [-p] [-t] [-l LAUNCHPAD_FILE] [-a] [-u UUID] [-d]

Execute a program in one of three different modes

optional arguments:
  -h, --help            show this help message and exit
  -f MODEL_FILE, --model_file MODEL_FILE
                        path to model file
  -g GRAMMAR_PATH, --grammar_path GRAMMAR_PATH
                        path to grammar root file
  -m {instant,deferred,workflow}, --mode {instant,deferred,workflow}
                        execution mode
  -p, --print-all-vars  prints all variables
  -t, --textx-debug     enable textx debug mode
  -l LAUNCHPAD_FILE, --launchpad_file LAUNCHPAD_FILE
                        path to launchpad file (workflow mode only)
  -a, --autorun         autorun the model (workflow mode only)
  -u UUID, --uuid UUID  UUID of a model to extend/run (workflow mode only)
  -d, --on-demand       run on demand (workflow mode only)
```

A persistent model can be retrieved by using the flag `-u UUID, --uuid UUID`. It can be optionally eveluated by adding the `-a, --autorun` and `-d, --on-demand` flags or/and extended by specifying a model extension with the flag `-f MODEL_FILE, --model_file MODEL_FILE`.

### Interactive session

The interactive session script `run_session.py` is an interactive tool managing a session for processing a model in *workflow mode*. Some of the startup flags are the same as for the script `run_model.py`. The flag `-a, --async-run` activates background evaluation using an workflow engine with a launcher thread. For this feature the [vre-middleware package](https://git.scc.kit.edu/virtmat-tools/vre-middleware) must be installed.


```
python scripts/run_session.py --help
usage: run_session.py [-h] [-c CONFIG_DIR] [-l LAUNCHPAD_FILE] [-g GRAMMAR_PATH] [-r] [-u UUID] [-d] [-a]

Execute a program remotely in deferred mode

optional arguments:
  -h, --help            show this help message and exit
  -c CONFIG_DIR, --config_dir CONFIG_DIR
                        path to configuration file (if -l unspecified)
  -l LAUNCHPAD_FILE, --launchpad_file LAUNCHPAD_FILE
                        path to launchpad file
  -g GRAMMAR_PATH, --grammar_path GRAMMAR_PATH
                        path to grammar root file
  -r, --autorun         run the model
  -u UUID, --uuid UUID  UUID of a model
  -d, --on-demand       run model nodes on demand
  -a, --async-run       run model in the background
```

#### Session behavior

In the session, inputs are interactively typed and the outputs are shown immediately. An empty model instance is created at startup and with every input the model instance is extended with the model source extension typed as input. The editing capabilities of the input fields are very limited.

#### Specific features

In the interactive session, if a valid parameter is typed it is interpreted as a `print` statement. For example, a typed input `a` is changed to `print(a)` before parsing. Multiline comments are not supported in the interactive session.

A successfully parsed and interpreted input cannot be modified. It becomes permanently part of the model source.

The session manager accepts additional inputs that are no model sources in SCL or AMML. These are sometimes called magic commands (like in iPython and Jupyter) and begin with the symbol `%`.

magic commands                     | action
-----------------------------------|-------- 
`%exit`, `%bye`, `%close`, `%quit` | close (quit) the session
`%start`                           | start the background launcher thread
`%stop`                            | stop the background launcher thread
`%uuid <UUID>`                     | create a new session with UUID
`%hist`, `%history`                | prints the current model source

The `%history` command prints all statements (without `print` statements that are not persisted), one per line, with timestamp of most recent evaluation update and current evaluation state. The evaluation state is currently passed through from the FireWorks workflow management system and is described in detail [here](https://materialsproject.github.io/fireworks/reference.html#fireworks-states). 


## Using the Jupyter kernel

The Jupyter kernel is still work in progress. Currently, some help can be obtained in [this README document](../virtmat_kernel/README.rst).


## Interpreting the error messages

Either the parser or the interpreter may detect errors in the program. Then they issue specific error messages in the following format: 

```
Error type: /absolute/path/to/model_source_file.vm:line:column --> context <--
Detailed error description
```

The `--> context <--` is only printed for syntax errors. If the Jupyter kernel or `run_session.py` is used then `None` is printed instead of a filename.

**NOTE:** If an error message is not in this particular format, especially if it is a Python exception with a stack trace beginning with the word "Traceback", then it is most likely due to a bug in the interpreter and not due to an error in the model source. In such cases please [submit an issue](https://git.scc.kit.edu/virtmat-tools/vre-language/-/issues/new) with the tag *Bug* providing a minimal input to reproduce the error and the error message.

### Simple examples of errors

Input:
```
b = 1 a = 2
```
Error:
```
Syntax error: None:1:7 --> b = 1 *a = 2  <--
Expected '[' or '^|;|\,|\)|$' or '**' or '*' or '/' or '+' or '-' or '<=' or '>=' or '>' or '<' or '!=' or '==' or 'if' or 'for' or 'on' or '^|;' or EOF
```

All syntax errors are issued by the parser.

Input:
```
b = a
```
Error:
```
Unknown object: None:1:5
Unknown object "a" of class "GeneralVariable"
```

Unknown object error is a semantic error issued by the parser when it cannot resolve a reference.

Input:
```
s = 1
s = 'hello'
```
Error:
```
Initialization error: None:2:1
Repeated initialization of "s"
```

Repeated initialization error is issued by the interpreter when it applies semantic constraints before the beginning of evaluation.