"""main script for model execution in one of three different modes"""
import os
from argparse import ArgumentParser
from textx import metamodel_from_file
from textx import get_children_of_type
from fireworks.fw_config import LAUNCHPAD_LOC
from fireworks.scripts.lpad_run import get_lp
from virtmat.language.metamodel.properties import add_properties
from virtmat.language.metamodel.processors import add_processors
from virtmat.language.interpreter.workflow_executor import NonCompletedException
from virtmat.language.interpreter.session_manager import Session
from virtmat.language.utilities.errors import error_handler

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
GRAMMAR_PATH = os.path.join(ROOT_DIR, '..', 'grammar', 'virtmat.tx')


def parse_clargs():
    """parse the command line arguments for the script"""
    m_description = 'Execute a program in one of three different modes'
    parser = ArgumentParser(description=m_description)
    parser.add_argument('-f', '--model_file', required=True,
                        help='path to model file')
    parser.add_argument('-g', '--grammar_path', required=False,
                        help='path to grammar root file', default=GRAMMAR_PATH)
    parser.add_argument('-m', '--mode', required=False, default='instant',
                        help='execution mode', choices=['instant', 'deferred', 'workflow'])
    parser.add_argument('-p', '--print-all-vars', default=False, required=False,
                        action='store_true', help='prints all variables')
    parser.add_argument('-t', '--textx-debug', default=False, required=False,
                        action='store_true', help='enable textx debug mode')
    parser.add_argument('-l', '--launchpad_file', default=LAUNCHPAD_LOC,
                        help='path to launchpad file (workflow mode only)')
    parser.add_argument('-a', '--autorun', default=False, required=False,
                        action='store_true', help='autorun the model (workflow mode only)')
    parser.add_argument('-u', '--uuid', default=None, required=False,
                        help='UUID of a model to extend/run (workflow mode only)')
    parser.add_argument('-d', '--on-demand', default=False, required=False,
                        action='store_true', help='run on demand (workflow mode only)')
    return parser.parse_args()


@error_handler
def run_instant_deferred(clargs, deferred=False):
    """run a model in instant or deferred mode"""

    meta = metamodel_from_file(clargs.grammar_path, auto_init_attributes=False,
                               debug=clargs.textx_debug)
    add_properties(meta, deferred_mode=deferred)
    add_processors(meta, constr_processors=True)

    with open(clargs.model_file, 'r', encoding='utf-8') as inp:
        model_str = inp.read()
    return meta.model_from_file(clargs.model_file, deferred_mode=deferred,
                                source_code=model_str, debug=clargs.textx_debug)


@error_handler
def run_workflow(clargs):
    """run a model through a workflow system"""
    session = Session(get_lp(clargs), uuid=clargs.uuid, grammar_path=clargs.grammar_path,
                      model_path=clargs.model_file, autorun=clargs.autorun,
                      on_demand=clargs.on_demand, textx_debug=clargs.textx_debug)
    prog = session.get_model()
    if prog:
        print('program UUID:', prog.uuid)
    return prog


@error_handler
def evaluate_prog(clargs, prog):
    """evaluate the model"""
    print(f'program output: >>>\n{prog.value}\n<<<')
    if clargs.print_all_vars:
        for var in get_children_of_type('Variable', prog):
            try:
                print(f'var: {var.name}, type: {var.type_}, value: {var.value}')
            except NonCompletedException:
                print(f'var: {var.name}, type: {var.type_}, value: n.c.')


def main():
    """main function"""
    clargs = parse_clargs()
    if clargs.mode == 'instant':
        prog = run_instant_deferred(clargs, deferred=False)
    elif clargs.mode == 'deferred':
        prog = run_instant_deferred(clargs, deferred=True)
    elif clargs.mode == 'workflow':
        prog = run_workflow(clargs)
    if prog is not None:
        evaluate_prog(clargs, prog)


if __name__ == '__main__':
    main()
