"""
check for cyclic dependencies
"""
from textx import get_children_of_type
from textx import get_location
from textx import textx_isinstance
from textx.exceptions import TextXError


class CyclicDependencyError(TextXError):
    """raise this exception if a cyclic dependency is detected"""
    err_type = 'Cyclic dependency'

    def __init__(self, var, ref):
        msgs = [f'variable: {var.name}, loc: {repr(get_location(var))}']
        msgs.append(f'reference: {repr(get_location(ref))}')
        message = self.err_type + ':\n' + '\n'.join(msgs)
        super().__init__(message, **get_location(ref), err_type=self.err_type)


def detect_cycles(lhs_var, rhs_var, meta):
    """detect cycles by a recursive walk through the dependency tree"""
    # objects that include references to Variable objects:
    obj_refs = {('VarReference', 'ref'), ('ChemTerm', 'species'),
                ('AMMLProperty', 'struct'), ('AMMLProperty', 'calc'),
                ('AMMLProperty', 'constrs')}
    objs = []
    for obj, attr in obj_refs:
        for var_ref_obj in get_children_of_type(obj, rhs_var.parameter):
            ref = getattr(var_ref_obj, attr)
            if textx_isinstance(ref, meta['Variable']):
                if isinstance(ref, list):
                    objs.extend([(var_ref_obj, r) for r in ref])
                else:
                    objs.append((var_ref_obj, ref))
    rhs_vars = set()
    for obj, var in objs:
        if var is not lhs_var:
            rhs_vars.add(var)
            rhs_vars.update(detect_cycles(lhs_var, var, meta))
        else:
            raise CyclicDependencyError(lhs_var, obj)
    return rhs_vars


def check_cycles_processor(model, metamodel):
    """model processor to check for cyclic dependencies in the program"""
    for var in get_children_of_type('Variable', model):
        detect_cycles(var, var, metamodel)
