"""
constraints applied to callable objects
"""
from textx import get_children_of_type
from textx import get_location
from textx import textx_isinstance
from textx.exceptions import TextXError
from .imports import check_function_import


def check_function_definition(obj, metamodel):
    """check the arguments matching in FunctionDefinition"""
    assert textx_isinstance(obj, metamodel['FunctionDefinition'])
    vrefs = get_children_of_type('VarReference', obj.expr)
    refs = [r.ref for r in vrefs if textx_isinstance(r.ref, metamodel['Dummy'])]
    refs_uniq_names = set(ref.name for ref in refs if ref in obj.args)
    args_names = [a.name for a in obj.args]
    args_uniq_names = set(args_names)
    if len(args_names) != len(args_uniq_names):
        err_type = 'Duplicate argument'
        args_names = [a.name for a in obj.args]
        args_uniq_names = set(args_names)
        dups = ', '.join(a for a in args_uniq_names if args_names.count(a) > 1)
        message = f'duplicate argument(s) in function {obj.name}: {dups}'
        raise TextXError(message, **get_location(obj), err_type=err_type)
    calls = get_children_of_type('FunctionCall', obj.expr)
    call_dummies = set()
    for call in calls:
        if not textx_isinstance(call.function, metamodel['ObjectImport']):
            for arg in call.function.args:
                call_dummies.add(arg.name)
    assert not any(d in args_uniq_names for d in call_dummies)
    args_uniq_names.update(call_dummies)
    refs_uniq_names.update(call_dummies)
    if list(sorted(refs_uniq_names)) != list(sorted(args_uniq_names)):
        err_type = 'Argument mismatch'
        message = (f'dummy variables {list(sorted(refs_uniq_names))} '
                   f'do not match arguments {list(sorted(args_uniq_names))} '
                   f'in function {obj.name}')
        raise TextXError(message, **get_location(obj), err_type=err_type)


def check_function_call(obj, metamodel):
    """check a function call object"""
    if textx_isinstance(obj.function, metamodel['ObjectImport']):
        check_function_import(obj.function)  # has a valid imported function
    else:
        assert textx_isinstance(obj.function, metamodel['FunctionDefinition'])
        if len(obj.params) != len(obj.function.args):
            err_type = 'Function parameter mismatch'
            message = (f'{obj.function.name}() takes {len(obj.function.args)} '
                       f'parameters but {len(obj.params)} were given')
            raise TextXError(message, **get_location(obj), err_type=err_type)


def check_functions_processor(model, metamodel):
    """processor to check function definitions and function calls"""
    for func in get_children_of_type('FunctionDefinition', model):
        check_function_definition(func, metamodel)
    for call in get_children_of_type('FunctionCall', model):
        check_function_call(call, metamodel)
