"""
contraints applied to imported objects
"""
from textx import get_children_of_type
from textx import get_location
from textx import get_metamodel
from textx import textx_isinstance
from textx.exceptions import TextXError


class ObjectImportError(TextXError):
    """raise this exception if an object cannot be imported"""
    def __init__(self, obj):
        err_type = 'Import error'
        obj_namespace = '.'.join(obj.namespace)
        message = f'Object could not be imported: {obj_namespace}.{obj.name}'
        super().__init__(message, **get_location(obj), err_type=err_type)


class NonCallableImportError(TextXError):
    """raise this exception if an imported object is not callable"""
    def __init__(self, obj):
        err_type = 'Non-callable import'
        obj_namespace = '.'.join(obj.namespace)
        message = f'Imported object is not callable: {obj_namespace}.{obj.name}'
        super().__init__(message, **get_location(obj), err_type=err_type)


def get_object_import(obj):
    """return an imported object"""
    assert hasattr(obj, 'name') and hasattr(obj, 'namespace')
    try:
        module = __import__('.'.join(obj.namespace), fromlist=[obj.name], level=0)
    except ImportError as err:
        raise ObjectImportError(obj) from err
    try:
        retval = getattr(module, obj.name)
    except AttributeError as err:
        raise ObjectImportError(obj) from err
    return retval


def check_imports_processor(model, _):
    """check all imported objects in a program"""
    for obj in get_children_of_type('ObjectImport', model):
        get_object_import(obj)


def check_function_import(obj):
    """check that an imported object is callable"""
    assert textx_isinstance(obj, get_metamodel(obj)['ObjectImport'])
    if not callable(get_object_import(obj)):
        raise NonCallableImportError(obj)
