# pylint: disable=protected-access
"""
static type checker based on a simple recursion
"""
from functools import cached_property
from textx import get_location
from textx import get_children_of_type, get_parent_of_type
from textx import get_metamodel
from textx import textx_isinstance
from textx.exceptions import TextXError
import pandas
import ase
from virtmat.language.utilities.amml import get_params_units, get_calculator_class
from virtmat.language.utilities.amml import get_params_magnitudes
from virtmat.language.metamodel.function import subst
from virtmat.language.utilities.textx import isinstance_m, isinstance_r
from virtmat.language.utilities.textx import get_reference
from virtmat.language.utilities.errors import textxerror_wrap
from virtmat.language.utilities.types import typemap, DType
from .imports import get_object_import, check_function_import
from .units import check_units


class ExpressionTypeError(TextXError):
    """raise this exception if an invalid type is used in an expression"""
    err_type = 'Expression type error'

    def __init__(self, obj):
        self.obj = obj
        message = 'invalid type in ' + obj.__class__.__name__
        super().__init__(message, **get_location(obj), err_type=self.err_type)


class TypeMismatchError(TextXError):
    """raise this exception if two objects have incompatible types"""
    err_type = 'Type mismatch error'

    def __init__(self, obj1, obj2):
        message = ('Type mismatch: \n' +
                   repr(obj1.type_) + repr(get_location(obj1)) + '\n' +
                   repr(obj2.type_) + repr(get_location(obj2)))
        super().__init__(message, **get_location(obj1), err_type=self.err_type)


class IterablePropertyError(TextXError):
    """raise this exception if an iterable does not have a property"""
    err_type = 'Iterable property error'

    def __init__(self, obj, obj_type, prop):
        message = f'{obj_type} object has no property "{prop}"'
        super().__init__(message, **get_location(obj), err_type=self.err_type)


def check_types_processor(model, _):
    """evaluate type of all objects that have type"""
    classes = ['VarTuple', 'Variable', 'Print', 'ConditionIn', 'ConditionComparison']
    for cls in classes:
        for obj in get_children_of_type(cls, model):
            _ = obj.type_


@textxerror_wrap
def check_series_type(obj, dtype=None):
    """check if the object type is series"""
    if not issubclass(obj.type_, typemap['Series']):
        raise TypeError('Object must be series or reference to series')
    if dtype is not None and obj.type_.datatype is not None:
        if isinstance(dtype, tuple):
            type_name = ' or '.join(t.__name__ for t in dtype)
        else:
            type_name = dtype.__name__
        obj_name = getattr(obj, 'name', 'object')
        msg = f'Series \"{obj_name}\" must be of type {type_name}'
        if isinstance(dtype, tuple):
            if not any(issubclass(obj.type_.datatype, t) for t in dtype):
                raise TypeError(msg)
        if not issubclass(obj.type_.datatype, dtype):
            raise TypeError(msg)
        if issubclass(obj.type_.datatype, bool):
            if ((isinstance(dtype, tuple) and bool not in dtype)
               or issubclass(dtype, bool)):
                raise TypeError(msg)


def check_quantity_type(obj):
    """check if the object type is quantity"""
    otype = obj.type_
    assert issubclass(otype, typemap['Number'])
    assert otype.datatype is None or issubclass(otype.datatype, (float, int))


def print_type(self):
    """check types in a print statement"""
    for par in self.params:
        _ = par.type_
    return str


def variable_type(self):
    """evaluate variable type"""
    if isinstance_m(self.parent, ['VarTuple']):
        vars_ = self.parent.variables
        datatypes = self.parent.parameter.type_.datatype
        return next(dt for var, dt in zip(vars_, datatypes) if var is self)
    return self.parameter.type_


def expression_type(self):
    """evaluate expression type"""
    try:
        for operand in self.operands:
            if operand.type_ is not None:
                check_quantity_type(operand)
    except AssertionError as err:
        raise ExpressionTypeError(self) from err
    datatypes = set(o.type_.datatype for o in self.operands if o.type_ is not None)
    if None in datatypes:
        datatypes.discard(None)
        if any(issubclass(t, float) for t in datatypes):
            datatype = float
        else:
            datatype = None
    elif any(issubclass(t, float) for t in datatypes):
        datatype = float
    else:
        assert all(issubclass(t, int) for t in datatypes)
        datatype = int
    return DType('Number', (typemap['Number'],), {'datatype': datatype})


def term_type(self):
    """evaluate term type"""
    try:
        for operand in self.operands:
            if operand.type_ is not None:
                check_quantity_type(operand)
    except AssertionError as err:
        raise ExpressionTypeError(self) from err
    if '/' in self.operators:
        datatype = float
    else:
        datatypes = set(o.type_.datatype for o in self.operands if o.type_ is not None)
        if None in datatypes:
            datatypes.discard(None)
            if all(issubclass(t, int) for t in datatypes):
                datatype = None
            else:
                datatype = float
        elif all(issubclass(t, int) for t in datatypes):
            datatype = int
        else:
            datatype = float
    return DType('Number', (typemap['Number'],), {'datatype': datatype})


def factor_type(self):
    """evaluate factor type"""
    try:
        for operand in self.operands:
            if operand.type_ is not None:
                check_quantity_type(operand)
    except AssertionError as err:
        raise ExpressionTypeError(self) from err
    datatypes = set(o.type_.datatype for o in self.operands if o.type_ is not None)
    if None in datatypes:
        datatypes.discard(None)
        if all(issubclass(t, int) for t in datatypes):
            datatype = None
        else:
            assert any(issubclass(t, float) for t in datatypes)
            datatype = float
    elif all(issubclass(t, int) for t in datatypes):
        datatype = int
    else:
        assert any(issubclass(t, float) for t in datatypes)
        datatype = float
    return DType('Number', (typemap['Number'],), {'datatype': datatype})


def power_type(self):
    """evaluate power type"""
    return self.operand.type_


def operand_type(self):
    """evaluate operand type"""
    return self.operand.type_


def boolean_expression_type(self):
    """evaluate boolean expression type"""
    try:
        if hasattr(self, 'operand'):
            assert issubclass(self.operand.type_, bool)
        else:
            assert all(issubclass(o.type_, bool) for o in self.operands)
    except AssertionError as err:
        raise ExpressionTypeError(self) from err
    return bool


def if_expression_type(self):
    """evaluate if-expression type"""
    if not issubclass(self.expr.type_, bool):
        raise ExpressionTypeError(self)
    if self.true_.type_ != self.false_.type_:
        raise TypeMismatchError(self.true_, self.false_)
    return self.true_.type_


def comparison_type(self):
    """evaluate comparison type"""
    if self.left.type_ is None or issubclass(self.left.type_, typemap['Number']):
        if self.right.type_ is not None and not issubclass(self.right.type_, typemap['Number']):
            raise TypeMismatchError(self.left, self.right)
        if self.left.type_ != self.left.type_:
            raise TypeMismatchError(self.left, self.right)
    elif issubclass(self.left.type_, (bool, str)):
        if self.left.type_ != self.right.type_:
            raise TypeMismatchError(self.left, self.right)
    return bool


def reference_type(self):
    """evaluate the type of a referenced object"""
    return self.ref.type_


def object_import_type(self):
    """evaluate the type of an imported object"""
    obj = get_object_import(self)
    assert not callable(obj)
    obj_type = type(obj)
    if issubclass(obj_type, (bool, str)):
        return obj_type
    if issubclass(obj_type, (int, float)):
        return DType('Number', (typemap['Number'],), {'datatype': obj_type})
    err_type = 'Import type error'
    msg = f'Unsupported type of imported object {obj_type}'
    raise TextXError(msg, **get_location(self), err_type=err_type)


def function_call_type(self):
    """evaluate the type of a function call"""
    metamodel = get_metamodel(self)
    if textx_isinstance(self.function, metamodel['ObjectImport']):
        if textx_isinstance(self.parent, metamodel['Operand']):
            ret_type = DType('Number', (typemap['Number'],), {})
        elif textx_isinstance(self.parent, metamodel['BooleanOperand']):
            ret_type = bool
        elif isinstance_m(self.parent, ['Comparison']):
            other = self.parent.right if self.parent.left is self else self.parent.left
            if not ((textx_isinstance(other, metamodel['FunctionCall']) and
                     textx_isinstance(other.function, metamodel['ObjectImport'])) or
                    (textx_isinstance(other, metamodel['VarReference']) and
                     textx_isinstance(other.ref, metamodel['ObjectImport']))):
                if issubclass(other.type_, typemap['Number']):
                    ret_type = DType('Number', (typemap['Number'],), {})
                else:
                    ret_type = other.type_
            else:
                ret_type = None
        elif isinstance_m(self.parent, ['IfFunction', 'IfExpression']):
            other = self.parent.true_ if self.parent.false_ is self else self.parent.false_
            if not ((textx_isinstance(other, metamodel['FunctionCall']) and
                     textx_isinstance(other.function, metamodel['ObjectImport'])) or
                    (textx_isinstance(other, metamodel['VarReference']) and
                     textx_isinstance(other.ref, metamodel['ObjectImport']))):
                ret_type = other.type_
            else:
                ret_type = None
        elif textx_isinstance(self.parent, metamodel['VarTuple']):
            typespec = {'datatype': tuple([None]*len(self.parent.variables))}
            ret_type = DType('Tuple', (tuple,), typespec)
        else:
            ret_type = None
    elif textx_isinstance(self.function, metamodel['FunctionDefinition']):
        try:
            ret_type = self.expr.type_
        except ExpressionTypeError as err:
            err_type = 'Function call parameter type error'
            msg = (f'Invalid type at {repr(get_location(err.obj))}'
                   f' in function call at {repr(get_location(self))}')
            raise TextXError(msg, **get_location(self), err_type=err_type) from err
    return ret_type


def tuple_type(self):
    """evaluate the types in a tuple object"""
    datatype = tuple(param.type_ for param in self.params)
    return DType('Tuple', (tuple,), {'datatype': datatype})


def var_tuple_type(self):
    """evaluate / check the types in a variables tuple object"""
    if self.parameter.type_ is None:
        return DType('Tuple', (tuple,), {})
    if (not issubclass(self.parameter.type_, tuple) or
            len(self.variables) != len(self.parameter.type_.datatype)):
        err_type = 'Tuple parameter mismatch'
        message = f'Expected {len(self.variables)} variable parameters'
        if issubclass(self.parameter.type_, tuple):
            message += f' but {len(self.parameter.type_.datatype)} were given'
        raise TextXError(message, **get_location(self), err_type=err_type)
    return self.parameter.type_


def series_type(self):
    """evaluate / check the type of a series"""
    if self.need_input:
        return DType('Series', (typemap['Series'],), {})
    types = []
    for elem in self.elements:
        if hasattr(elem, 'type_'):
            if isinstance(elem.type_, DType):
                types.append(elem.type_)
            else:
                assert issubclass(elem.type_, (bool, str))
                types.append(elem.type_)
        elif isinstance(elem, (int, float)):
            types.append(type(elem))
        else:
            assert isinstance_m(elem, ['Null'])
    if len(set(types)) > 1:
        err_type = 'Series type error'
        msg = (f'Series elements must have one type but {len(set(types))}'
               f' types were found')
        raise TextXError(msg, **get_location(self), err_type=err_type)
    datatype = next(iter(types)) if len(types) > 0 else None
    typespec = {'datatype': datatype, 'datalen': len(self.elements)}
    return DType('Series', (typemap['Series'],), typespec)


def table_type(self):
    """evaluate / check the types in a table object"""
    if self.need_input:
        return DType('Table', (typemap['Table'],), {})
    names = [c.name for c in self.columns if isinstance_m(c, ['Series'])]
    if len(set(names)) != len(names):
        err_type = 'Table column names error'
        msg = 'Repeating column names were found in table'
        raise TextXError(msg, **get_location(self), err_type=err_type)
    if not all(issubclass(c.type_, typemap['Series']) for c in self.columns):
        err_type = 'Table column type error'
        elem = next(c for c in self.columns if not issubclass(c.type_, typemap['Series']))
        msg = 'The type of table column must be series'
        raise TextXError(msg, **get_location(elem), err_type=err_type)
    types = tuple(c.type_ for c in self.columns)
    datatypes = tuple(t.datatype for t in types)
    datalens = set(t.datalen for t in types)
    if len(datalens) != 1:
        err_type = 'Table column size error'
        msg = f'Table columns must have one size but {len(datalens)} sizes were found'
        raise TextXError(msg, **get_location(self), err_type=err_type)
    type_spec = {'datatype': datatypes, 'datalen': next(iter(datalens))}
    return DType('Table', (typemap['Table'],), type_spec)


def array_type(self):
    """return array type"""
    meta = get_metamodel(self)
    mtypes = ['BoolArray', 'StrArray', 'IntArray', 'FloatArray']
    dtypes = [bool, str, int, float]
    mtype, dtype = next((m, d) for m, d in zip(mtypes, dtypes)
                        if textx_isinstance(self, meta[m]))
    return DType(mtype, (typemap[mtype],), {'datatype': dtype})


def get_array_type(datatype, typespec):
    """construct and return the proper array type depending on datatype"""
    mtypes = ['BoolArray', 'StrArray', 'IntArray', 'FloatArray']
    dtypes = [bool, str, int, float]
    mtype = next(m for m, d in zip(mtypes, dtypes) if issubclass(datatype, d))
    return DType(mtype, (typemap[mtype],), typespec)


def get_sliced_type(obj):
    """return a type slice of an iterable object"""
    type_ = obj.obj.type_.datatype
    size_ = obj.obj.type_.datalen
    name_ = obj.obj.type_.__name__
    if obj.slice:
        obj_slice = slice(obj.start, obj.stop, obj.step)
        if size_ is not None:
            try:
                slice_size = len(range(*obj_slice.indices(size_)))
            except ValueError as err:
                if 'step cannot be zero' in str(err):
                    raise TextXError(str(err).capitalize(), **get_location(obj),
                                     err_type='Slice step error') from err
                raise err
            typespec = {'datatype': type_, 'datalen': slice_size}
        else:
            typespec = {'datatype': type_, 'datalen': size_}
        if obj.array:
            ret = get_array_type(type_, typespec)
        else:
            ret = DType(name_, (obj.obj.type_,), typespec)
    else:
        typespec = {'datatype': type_, 'datalen': size_}
        if obj.array:
            ret = get_array_type(type_, typespec)
        else:
            ret = DType(name_, (obj.obj.type_,), typespec)
    return ret


def get_property_type(var, type_, accessor):
    """return the type of a property using an accessor"""
    if type_ is None:
        return None
    if accessor.id is not None:
        if issubclass(type_, typemap['Table']):
            if isinstance_m(var.parameter, ['Table']):  # table literal
                series = var.parameter.get_column(accessor.id)
                if series is None:
                    msg = f'column {accessor.id} not found in Table {var.name}'
                    raise KeyError(msg)
                rettype = series.type_
            else:
                rettype = DType('Series', (typemap['Series'],), {'datatype': None})
        elif issubclass(type_, typemap['Series']):
            raise ValueError('Invalid use of series name')
        elif issubclass(type_, typemap['AMMLStructure']):
            if accessor.id == 'name':
                rettype = str
            elif accessor.id == 'atoms':
                typespec_0 = {'datatype': (str, float, float, float)}
                datatype_0 = DType('Table', (typemap['Table'],), typespec_0)
                typespec_1 = {'datatype': datatype_0}
                rettype = DType('Series', (typemap['Series'],), typespec_1)
            elif accessor.id == 'pbc':
                typespec = {'datatype': typemap['BoolArray']}
                rettype = DType('Series', (typemap['Series'],), typespec)
            elif accessor.id == 'cell':
                typespec_0 = {'datatype': typemap['Number']}
                datatype_0 = DType('FloatArray', (typemap['FloatArray'], ), typespec_0)
                typespec_1 = {'datatype': datatype_0}
                rettype = DType('Series', (typemap['Series'],), typespec_1)
            else:
                raise KeyError(f'Invalid key \"{accessor.id}\" in {type_.__name__}')
        elif issubclass(type_, typemap['AMMLCalculator']):
            if accessor.id in ('name', 'pinning', 'version'):
                rettype = str
            elif accessor.id == 'parameters':
                rettype = DType('Table', (typemap['Table'],), {'datatype': None})
            else:
                raise KeyError(f'Invalid key \"{accessor.id}\" in {type_.__name__}')
        elif issubclass(type_, typemap['AMMLProperty']):
            if accessor.id == 'names':
                rettype = DType('Tuple', (typemap['Tuple'],), {'datatype': None})
            elif accessor.id == 'calc':
                rettype = DType('AMMLCalculator', (typemap['AMMLCalculator'],), {})
            elif accessor.id == 'struct':
                rettype = DType('AMMLStructure', (typemap['AMMLStructure'],), {})
            elif accessor.id == 'output_structure':
                datatype = DType('AMMLStructure', (typemap['AMMLStructure'],), {})
                rettype = DType('Series', (typemap['Series'],), {'datatype': datatype})
            elif accessor.id in ('forces', 'dipole'):
                typespec_0 = {'datatype': float}
                datatype_0 = DType('FloatArray', (typemap['FloatArray'], ), typespec_0)
                rettype = DType('Series', (typemap['Series'],), {'datatype': datatype_0})
            elif accessor.id == 'vibrational_energies':
                typespec_0 = {'datatype': float}
                datatype_0 = DType('Series', (typemap['Series'], ), typespec_0)
                rettype = DType('Series', (typemap['Series'],), {'datatype': datatype_0})
            elif accessor.id == 'energy_minimum':
                typespec = {'datatype': bool}
                rettype = DType('Series', (typemap['Series'],), typespec)
            elif accessor.id == 'energy':
                typespec = {'datatype': float}
                rettype = DType('Series', (typemap['Series'],), typespec)
            elif accessor.id == 'constrs':
                datatype = DType('AMMLConstraint', (typemap['AMMLConstraint'],), {})
                rettype = DType('Tuple', (typemap['Tuple'],), {'datatype': datatype})
            elif accessor.id == 'task':
                rettype = str
            elif accessor.id == 'results':
                rettype = DType('Table', (typemap['Table'],), {'datatype': None})
            else:
                raise KeyError(f'Invalid key \"{accessor.id}\" in {type_.__name__}')
        elif issubclass(type_, typemap['ChemSpecies']):
            keys = ['energy', 'enthalpy', 'entropy', 'free_energy', 'zpe', 'temperature']
            if accessor.id in keys:
                datatype = float
                rettype = DType('Series', (typemap['Series'],), {'datatype': datatype})
            elif accessor.id == 'name':
                rettype = str
            else:
                raise KeyError(f'Invalid key \"{accessor.id}\" in {type_.__name__}')
        elif issubclass(type_, typemap['ChemReaction']):
            keys = ['energy', 'enthalpy', 'entropy', 'free_energy', 'zpe', 'temperature']
            if accessor.id in keys:
                datatype = float
                rettype = DType('Series', (typemap['Series'],), {'datatype': datatype})
            else:
                raise KeyError(f'Invalid key \"{accessor.id}\" in {type_.__name__}')
        else:
            msg = f'Invalid use of an ID in type {type_.__name__}'
            raise TypeError(msg)
    else:
        assert accessor.index is not None
        if hasattr(var, '_tx_metamodel'):
            if isinstance_m(var.parameter, ['Table', 'Series']):
                if not abs(accessor.index) < type_.datalen:
                    raise IndexError('Index out of range')
        if issubclass(type_, typemap['Table']):
            rettype = DType('Tuple', (typemap['Tuple'],), {'datatype': type_.datatype})
        elif issubclass(type_, typemap['Series']):
            if issubclass(type_.datatype, (str, bool)):
                rettype = type_.datatype
            elif issubclass(type_.datatype, (int, float)):
                rettype = DType('Number', (typemap['Number'],), {'datatype': type_.datatype})
            else:
                rettype = type_.datatype
        elif issubclass(type_, typemap['AMMLStructure']):
            types = (typemap['Table'], typemap['FloatArray'], typemap['BoolArray'])
            rettype = DType('Tuple', (typemap['Tuple'],), {'datatype': types})
        elif issubclass(type_, typemap['AMMLCalculator']):
            rettype = DType('Tuple', (typemap['Tuple'],), {'datatype': None})
        elif issubclass(type_, typemap['AMMLProperty']):
            rettype = DType('Tuple', (typemap['Tuple'],), {'datatype': None})
        elif issubclass(type_, (typemap['ChemSpecies'], typemap['ChemReaction'])):
            rettype = DType('Tuple', (typemap['Tuple'],), {'datatype': typemap['Number']})
        else:
            msg = f'Invalid use of index in type {type_.__name__}'
            raise TypeError(msg)
    return rettype


@textxerror_wrap
def property_reference_type(self):
    """return the type of a reference to an object with data accessors"""
    type_ = self.varref.type_
    for accessor in self.accessors:
        type_ = get_property_type(self.varref.ref, type_, accessor)
    return type_


def iterable_property_type(self):
    """evaluate / check the type of an iterable property object"""
    if self.obj.type_ is None:
        ret_type = None
    elif issubclass(self.obj.type_, typemap['Table']):
        if self.name_:
            raise IterablePropertyError(self, 'Table', 'name')
        if self.columns and isinstance_m(self.obj.varref.ref.parameter, ['Table']):
            typespec = {'datatype': (str,)*len(self.obj.varref.ref.parameter.columns)}
            ret_type = DType('Tuple', (tuple,), typespec)
        elif self.array:
            raise IterablePropertyError(self, 'Table', 'array')
        else:
            ret_type = get_sliced_type(self)
    elif issubclass(self.obj.type_, typemap['Series']):
        if self.columns:
            raise IterablePropertyError(self, 'Series', 'columns')
        ret_type = str if self.name_ else get_sliced_type(self)
    else:
        if self.array:
            raise IterablePropertyError(self, 'Plain type / tuple', 'array')
        ret_type = self.obj.type_
    return ret_type


def iterable_query_type(self):
    """evaluate / check the type of an iterable query object"""
    classes = ['Table', 'Series', 'IterableQuery', 'IterableProperty']
    param = self.obj.varref.ref.parameter
    if isinstance_r(param, classes):
        if len(self.columns) > 0:
            if isinstance_m(param, ['Series']):
                raise IterablePropertyError(self, 'Series', 'columns')
            columns = [param.get_column(c) for c in self.columns]
            types = tuple(c.type_.datatype for c in columns)
        else:
            types = self.obj.type_.datatype
    else:
        types = None
    size = None if self.where else self.obj.type_.datalen
    typespec = {'datatype': types, 'datalen': size}
    return DType(self.obj.type_.__name__, (self.obj.type_,), typespec)


def check_column(pref, column):
    """check whether a PropertyReference pref contains a column (str)"""
    pref_par = pref.varref.ref.parameter
    pref_name = pref.varref.ref.name
    assert column is not None
    if (issubclass(pref.type_, typemap['Series']) and
       isinstance_m(pref_par, ['Series'])):
        if column != pref_par.name:
            msg = f'column \"{column}\" does not match Series name \"{pref_name}\"'
            raise ValueError(msg)
    if (issubclass(pref.type_, typemap['Table']) and
       isinstance_m(pref_par, ['Table'])):
        if pref_par.get_column(column) is None:
            msg = f'column \"{column}\" not found in Table \"{pref_name}\"'
            raise ValueError(msg)


@textxerror_wrap
def condition_in_type(self):
    """evaluate / check the type of condition in"""
    check_column(get_parent_of_type('IterableQuery', self).obj, self.column)
    return typemap['Series']


@textxerror_wrap
def condition_comparison_type(self):
    """evaluate / check the type of condition comparison"""
    prop_ref = get_parent_of_type('IterableQuery', self).obj
    if self.column_left is not None:
        check_column(prop_ref, self.column_left)
    if self.column_right is not None:
        check_column(prop_ref, self.column_right)
    return typemap['Series']


def range_type(self):
    """evaluate / check the type of the range builtin function"""
    for par in (self.start, self.stop, self.step):
        if (issubclass(par.type_, (bool, str))
           or not issubclass(par.type_.datatype, (int, float))):
            msg = f'Range paramter {par} must be numeric type'
            err_type = 'Range type error'
            raise TextXError(msg, **get_location(par), err_type=err_type)
    types = (self.start.type_, self.stop.type_, self.step.type_)
    if any(issubclass(t.datatype, float) for t in types):
        datatype = float
    else:
        assert all(issubclass(t.datatype, int) for t in types)
        datatype = int
    return DType('Series', (typemap['Series'],), {'datatype': datatype})


def sum_type(self):
    """evaluate / check the type of the sum builtin function"""
    try:
        if self.parameter is not None:
            check_series_type(self.parameter, (int, float))
            typespec = {'datatype': self.parameter.type_.datatype}
        elif self.params:
            assert all(issubclass(p.type_, typemap['Number']) for p in self.params)
            if any(p.type_.datatype is None for p in self.params):
                typespec = {'datatype': None}
            elif any(issubclass(p.type_.datatype, float) for p in self.params):
                typespec = {'datatype': float}
            else:
                typespec = {'datatype': int}
    except AssertionError as err:
        msg = 'Sum accepts only inputs of numerical types'
        raise TextXError(msg, **get_location(self), err_type='Sum type error') from err
    return DType('Number', (typemap['Number'],), typespec)


def map_type(self):
    """evaluate / check the type of the map builtin function"""
    meta = get_metamodel(self)
    func = self.lambda_ if self.lambda_ else self.function
    for param in self.params:
        check_series_type(param)
    dsize = self.params[0].type_.datalen
    if not all(p.type_.datalen == dsize or p.type_.datalen is None for p in self.params):
        msg = 'Map input series must have equal size'
        err_type = 'Map function error'
        raise TextXError(msg, **get_location(self), err_type=err_type)
    if not isinstance_m(func, ['ObjectImport']):
        if len(func.args) != len(self.params):
            msg = ('Number of map function arguments and number of input '
                   'series must be equal')
            err_type = 'Map function error'
            raise TextXError(msg, **get_location(self), err_type=err_type)
        params = []
        types = []
        for par in self.params:
            datatype = par.type_.datatype
            if datatype is not None:
                if issubclass(datatype, bool):
                    params.append(meta['Bool']())
                    types.append(bool)
                elif issubclass(datatype, str):
                    params.append(meta['String']())
                    types.append(str)
                elif issubclass(datatype, (int, float)):
                    param = meta['Number']()
                    param.inp_value = 0 if issubclass(datatype, int) else 0.
                    param.inp_units = None
                    param.need_input = False
                    params.append(param)
                    types.append(DType('Number', (typemap['Number'],), {'datatype': datatype}))
        if types and params:
            new_type = subst(self, func, params, types).type_
            if new_type is None or issubclass(new_type, (bool, str)):
                new_datatype = new_type
            else:
                new_datatype = new_type.datatype
            typespec = {'datatype': new_datatype, 'datalen': dsize}
        else:
            typespec = {}
    else:
        check_function_import(func)
        typespec = {'datalen': dsize}
    return DType('Series', (typemap['Series'],), typespec)


def filter_type(self):
    """evaluate / check the type of the filter builtin function"""
    func = self.lambda_ if self.lambda_ else self.function
    check_series_type(self.parameter)
    datatype = self.parameter.type_.datatype
    if not isinstance_m(func, ['ObjectImport']):
        if len(func.args) != 1:
            msg = 'Filter function must have only one argument'
            err_type = 'Filter function error'
            raise TextXError(msg, **get_location(self), err_type=err_type)
        if (isinstance_m(func.expr, ['FunctionCall'])
                and isinstance_m(func.expr.function, ['ObjectImport'])):
            check_function_import(func.expr.function)
        else:
            if datatype is not None:
                if issubclass(datatype, (bool, str, int, float)):
                    meta = get_metamodel(self)
                    if issubclass(datatype, bool):
                        param = meta['Bool']()
                        type_ = bool
                    elif issubclass(datatype, str):
                        param = meta['String']()
                        type_ = str
                    elif issubclass(datatype, (int, float)):
                        param = meta['Number']()
                        param.inp_value = 0 if issubclass(datatype, int) else 0.
                        param.inp_units = None
                        param.need_input = False
                        type_ = DType('Number', (typemap['Number'],), {'datatype': datatype})
                    if not issubclass(subst(self, func, [param], [type_]).type_, bool):
                        msg = 'Filter function must be of boolean type'
                        err_type = 'Filter function error'
                        raise TextXError(msg, **get_location(self), err_type=err_type)
    else:
        check_function_import(func)
    return DType('Series', (typemap['Series'],), {'datatype': datatype})


def reduce_type(self):
    """evaluate / check the type of the reduce builtin function"""
    meta = get_metamodel(self)
    func = self.lambda_ if self.lambda_ else self.function
    check_series_type(self.parameter)
    datatype = self.parameter.type_.datatype
    if not isinstance_m(func, ['ObjectImport']):
        if len(func.args) != 2:
            msg = 'Reduce function must have exactly two arguments'
            err_type = 'Reduce function error'
            raise TextXError(msg, **get_location(self), err_type=err_type)
        if (isinstance_m(func.expr, ['FunctionCall'])
                and isinstance_m(func.expr.function, ['ObjectImport'])):
            check_function_import(func.expr.function)
            ret_type = None
        else:
            if datatype is not None:
                if issubclass(datatype, (bool, str, int, float)):
                    if issubclass(datatype, bool):
                        param = meta['Bool']()
                        type_ = bool
                    elif issubclass(datatype, str):
                        param = meta['String']()
                        type_ = str
                    elif issubclass(datatype, (int, float)):
                        param = meta['Number']()
                        param.inp_value = 0 if issubclass(datatype, int) else 0.
                        param.inp_units = None
                        param.need_input = False
                        type_ = DType('Number', (typemap['Number'],), {'datatype': datatype})
                    ret_type = subst(self, func, [param]*2, [type_]*2).type_
                else:
                    ret_type = None
            else:
                ret_type = None
    else:
        check_function_import(func)
        ret_type = None
    return ret_type


def boolean_reduce_type(self):
    """evaluate / check the type of the boolean reduce builtin function"""
    try:
        if self.parameter is not None:
            check_series_type(self.parameter)
            if self.parameter.type_.datatype is not None:
                assert issubclass(self.parameter.type_.datatype, bool)
        else:
            assert all(issubclass(p.type_, bool) for p in self.params)
    except AssertionError as err:
        msg = 'Boolean reduce parameters must be of boolean type'
        err_type = 'Boolean reduce function error'
        raise TextXError(msg, **get_location(self), err_type=err_type) from err
    return bool


def in_type(self):
    """evaluate / check the type of in-expression"""
    if self.parameter is not None:
        check_series_type(self.parameter)
    return bool


def number_type(self):
    """evaluate the type of a number literal"""
    if self.inp_value is not None:
        typespec = {'datatype': type(self.inp_value)}
        return DType('Number', (typemap['Number'],), typespec)
    return DType('Number', (typemap['Number'],), {})


def amml_structure_type(self):
    """evaluate the type of structure object"""
    rtype = DType('AMMLStructure', (typemap['AMMLStructure'],), {})
    if self.filename or self.url:
        return rtype
    tab_typs = {'atoms': 'Table', 'cell': 'FloatArray', 'pbc': 'BoolArray'}
    if 'atoms' not in [c.name for c in self.tab.columns]:
        msg = '\'atoms\' missing in \'structure\''
        raise TextXError(msg, **get_location(self.tab), err_type='Value error')
    if not all(c.name in tab_typs for c in self.tab.columns):
        col = next(c for c in self.tab.columns if c.name not in tab_typs)
        msg = f'invalid object \'{col.name}\' in \'structure\''
        raise TextXError(msg, **get_location(col), err_type='Value error')

    for col in self.tab.columns:
        if isinstance_m(col, ['Series']):
            for elem in col.elements:
                if not hasattr(elem, '_tx_metamodel'):
                    msg = (f'element in \'{col.name}\' must have type '
                           f'{tab_typs[col.name]} but has type {type(elem)}')
                    raise TextXError(msg, **get_location(col), err_type='Type error')
                obj = get_reference(elem)
                if not isinstance_m(obj, [tab_typs[col.name]]):
                    msg = (f'element in \'{col.name}\' must have type '
                           f'{tab_typs[col.name]} but has type {obj.__class__.__name__}')
                    raise TextXError(msg, **get_location(obj), err_type='Type error')
        fquantity = (float, typemap['Number'])
        atoms_dtyps = {'symbols': str, 'x': fquantity, 'y': fquantity, 'z': fquantity,
                       'px': fquantity, 'py': fquantity, 'pz': fquantity, 'tags': int,
                       'masses': fquantity}

    atoms_col = self.tab.get_column('atoms')
    if atoms_col is not None:
        for atoms_elem in atoms_col.elements:
            atoms = get_reference(atoms_elem)
            assert issubclass(atoms.type_, typemap['Table'])
            for col in atoms.columns:
                assert issubclass(col.type_, typemap['Series'])
                if isinstance_m(col, ['Series']):
                    if col.name not in atoms_dtyps:
                        msg = f'invalid object \'{col.name}\' in \'atoms\''
                        raise TextXError(msg, **get_location(col), err_type='Value error')
                    if not issubclass(col.type_.datatype, atoms_dtyps[col.name]):
                        msg = (f'\'{col.name}\' must have type {atoms_dtyps[col.name]} but has'
                               f' type {col.type_.datatype}')
                        raise TextXError(msg, **get_location(col), err_type='Type error')

            symbols = atoms.get_column('symbols')
            if symbols is not None:
                for elem in symbols.elements:
                    if elem.value not in ase.data.chemical_symbols:
                        msg = f'invalid chemical symbol {elem.value} in \'symbols\''
                        raise TextXError(msg, **get_location(elem), err_type='Value error')

            momenta = ('px', 'py', 'pz')
            if any(atoms.get_column(p) is not None for p in momenta):
                if not all(atoms.get_column(p) is not None for p in momenta):
                    missed = next(p for p in momenta if atoms.get_column(p) is None)
                    msg = f'{missed} missing in \'atoms\''
                    raise TextXError(msg, **get_location(atoms), err_type='Value error')

            if self.tab.get_column('cell') is not None:
                for cell_elem in self.tab.get_column('cell').elements:
                    check_units(get_reference(cell_elem), '[length]')
            for column in ('x', 'y', 'z'):
                coord = atoms.get_column(column)
                if coord is not None:
                    check_units(coord, '[length]')
            if any(atoms.get_column(x) is not None for x in ('px', 'py', 'pz')):
                for column in ('px', 'py', 'pz'):
                    momentum = atoms.get_column(column)
                    if momentum is not None:
                        check_units(momentum, '[length]*[mass]/[time]')
            if atoms.get_column('masses') is not None:
                check_units(atoms.get_column('masses'), '[mass]')
    return rtype


@textxerror_wrap
def amml_calculator_type(self):
    """evaluate the type of calculator object"""
    rtype = DType('AMMLCalculator', (typemap['AMMLCalculator'],), {})
    if self.filename or self.url or self.parameters is None:
        return rtype
    literals = ['Number', 'Bool', 'String', 'FloatArray',
                'IntArray', 'BoolArray', 'StrArray']
    if isinstance_m(self.parameters, ['Table']):
        params_series = []
        for col in self.parameters.columns:
            if isinstance_m(col, ['Series']):
                if all(isinstance(e, (int, float)) for e in col.elements):
                    params_series.append(col.value)
                elif all(isinstance_m(e, literals) for e in col.elements):
                    params_series.append(col.value)
        if len(params_series) > 0:
            recs = pandas.concat(params_series, axis=1).to_dict(orient='records')
            for calc_params in recs:
                units = get_params_units(self.name, calc_params)
                calc_class = get_calculator_class(self.name)
                calc_class(**get_params_magnitudes(calc_params, units))
    return rtype


@textxerror_wrap
def amml_property_type(self):
    """evaluate the type of property object"""
    if not issubclass(self.struct.type_, typemap['AMMLStructure']):
        msg = f'Parameter \"{self.struct.name}\" must be an AMML structure'
        raise TextXError(msg, **get_location(self), err_type='Type error')
    if not issubclass(self.calc.type_, typemap['AMMLCalculator']):
        msg = f'parameter \"{self.calc.name}\" must be an AMML calculator'
        raise TextXError(msg, **get_location(self), err_type='Type error')
    if (isinstance_m(self.struct.parameter, ['AMMLStructure']) and
       self.struct.parameter.tab is not None):
        atoms = self.struct.parameter.tab.get_column('atoms')
        if atoms is not None:
            symbs = [e.get_column('symbols').elements for e in atoms.elements]
            for constr in self.constrs:
                if isinstance_m(constr.parameter, ['AMMLConstraint']):
                    if isinstance_m(constr.parameter.fixed, ['Series']):
                        elems = constr.parameter.fixed.elements
                        if not all(len(elems) == len(s) for s in symbs):
                            msg = ('The list of fixed/non-fixed atoms in constraints'
                                   ' and atoms in structure have different lengths')
                            raise ValueError(msg)
    return DType('AMMLProperty', (typemap['AMMLProperty'],), {})


def amml_constraint_type(self):
    """evaluate the type of constraint object"""
    if self.fixed is not None:
        if not (issubclass(self.fixed.type_, typemap['Series'])
           and issubclass(self.fixed.type_.datatype, bool)):
            msg = 'parameter must be a boolean series'
            raise TextXError(msg, **get_location(self.fixed), err_type='Type error')
    if self.direction is not None:
        if (not all(isinstance(e, int) for e in self.direction.elements)
           or len(self.direction.elements) != 3):
            msg = 'direction vector must be 1-dim integer array with 3 elements'
            raise TextXError(msg, **get_location(self.direction), err_type='Type error')
    return DType('AMMLConstraint', (typemap['AMMLConstraint'],), {})


def check_numerical_props(tab):
    """check if a table contains only series with given names of float type"""
    if tab is not None:
        if tab.columns:
            for prop in tab.columns:
                keys = ['energy', 'enthalpy', 'entropy', 'free_energy', 'zpe', 'temperature']
                if prop.name not in keys:
                    msg = f'invalid property \'{prop.name}\''
                    raise TextXError(msg, **get_location(prop), err_type='Value error')
                check_series_type(prop, (typemap['Number'], float))
        if tab.columns_tuple:
            for prop in tab.columns_tuple.params:
                check_series_type(prop, (typemap['Number'], float))


def chem_reaction_type(self):
    """evaluate the type of a chemical reaction object"""
    check_numerical_props(self.props)
    for term in self.educts + self.products:
        if not isinstance_m(term.species.parameter, ['ChemSpecies']):
            msg = f'{term.species.name} is no chemical species'
            raise TextXError(msg, **get_location(term), err_type='Type error')
    return DType('ChemReaction', (typemap['ChemReaction'],), {})


def chem_species_type(self):
    """evaluate the type of a chemical species object"""
    check_numerical_props(self.props)
    return DType('ChemSpecies', (typemap['ChemSpecies'],), {})


def add_type_properties(metamodel):
    """Add object class properties using monkey style patching"""
    mapping_dict = {
        'Print': print_type,
        'Variable': variable_type,
        'VarReference': reference_type,
        'Power': power_type,
        'Factor': factor_type,
        'Term': term_type,
        'Expression': expression_type,
        'Operand': operand_type,
        'BooleanOperand': operand_type,
        'And': boolean_expression_type,
        'Or': boolean_expression_type,
        'Not': boolean_expression_type,
        'IfFunction': if_expression_type,
        'IfExpression': if_expression_type,
        'Comparison': comparison_type,
        'ObjectImport': object_import_type,
        'FunctionCall': function_call_type,
        'Number': number_type,
        'Tuple': tuple_type,
        'VarTuple': var_tuple_type,
        'Series': series_type,
        'Table': table_type,
        'BoolArray': array_type,
        'StrArray': array_type,
        'IntArray': array_type,
        'FloatArray': array_type,
        'PropertyReference': property_reference_type,
        'IterableProperty': iterable_property_type,
        'IterableQuery': iterable_query_type,
        'ConditionIn': condition_in_type,
        'ConditionComparison': condition_comparison_type,
        'Range': range_type,
        'In': in_type,
        'Any': boolean_reduce_type,
        'All': boolean_reduce_type,
        'Sum': sum_type,
        'Map': map_type,
        'Filter': filter_type,
        'Reduce': reduce_type,
        'AMMLStructure': amml_structure_type,
        'AMMLCalculator': amml_calculator_type,
        'AMMLProperty': amml_property_type,
        'AMMLConstraint': amml_constraint_type,
        'ChemReaction': chem_reaction_type,
        'ChemSpecies': chem_species_type
    }
    for key, function in mapping_dict.items():
        metamodel[key].type_ = cached_property(function)
        metamodel[key].type_.__set_name__(metamodel[key], 'type_')

    metamodel['Dummy'].type_ = None
    metamodel['Bool'].type_ = bool
    metamodel['String'].type_ = str
    metamodel['Program'].type_ = str
    metamodel['ConditionNot'].type_ = typemap['Series']
    metamodel['ConditionAnd'].type_ = typemap['Series']
    metamodel['ConditionOr'].type_ = typemap['Series']
