# pylint: disable = protected-access, fixme
"""
Language interpreter for delayed evaluation

The methods below return the relevant Python functions together with a tuple
with their parameters. The evaluations are performed via delegation.

References cannot or should not be peformed via delegation. They are regarded as
function parameters, i.e. they constitute the list of inputs for the function
tasks (VarReference). Therefore, these elements have no func properties and the
checkref() function should be used instead.
"""
import logging
import pathlib
from itertools import islice, chain
from functools import partial, reduce, cached_property
from operator import gt, lt, eq, ne, le, ge
from operator import add, neg, mul, truediv
from operator import and_, or_, not_, invert
import numpy
import pandas
import pint_pandas
from textx import get_parent_of_type, get_metamodel
from fireworks.utilities.fw_serializers import load_object
from virtmat.language.constraints.imports import get_object_import
from virtmat.language.utilities.textx import isinstance_m
from virtmat.language.utilities.ioops import load_value
from virtmat.language.utilities.formatters import formatter
from virtmat.language.utilities.errors import textxerror_wrap
from virtmat.language.utilities.types import checktype, settype, typemap
from virtmat.language.utilities import amml, chemistry
from .instant_executor import program_value, plain_type_value
from .instant_executor import get_array_aslist

LOGLEVEL = logging.INFO
# LOGLEVEL = logging.DEBUG
logging.basicConfig(level=LOGLEVEL)

# binary operator--function map
binop_map = {'>': gt, '<': lt, '==': eq, '!=': ne, '<=': le, '>=': ge,
             '+': add, '-': lambda x, y: add(x, neg(y)), '*': mul, '/': truediv,
             'or': lambda x, y: x or y, 'and': lambda x, y: x and y,
             'or_': or_, 'and_': and_, '**': pow}


def checkref(obj):
    """return a 2-tuple containing a function and a list of parameters"""
    logging.debug('checkref:%s', repr(obj))
    if isinstance_m(obj, ['VarReference']):
        if isinstance_m(obj.ref, ['Variable', 'Dummy', 'ObjectImport']):
            # make sure the referenced obj has a name
            ret = (lambda x: x, (obj.ref,))
        else:
            ret = checkref(obj.ref)
    else:
        ret = obj.func
    return ret


def dummies_right(func, args):
    """reorder arguments so that dummy variables are the end of the list"""
    params = []
    order = []
    for index, arg in enumerate(args):
        if not isinstance_m(arg, ['Dummy']):
            params.append(arg)
            order.append(index)
    for index, arg in enumerate(args):
        if isinstance_m(arg, ['Dummy']):
            params.append(arg)
            order.append(index)
    return lambda *nargs: func(*[nargs[order[i]-1] for i in order]), tuple(params)


def reference_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    return checkref(self)


def print_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    func_pars = [checkref(par) for par in self.params]
    funcs = [t[0] for t in func_pars]
    pars = tuple(t[1] for t in func_pars)
    pars_lens = [len(p) for p in pars]

    def retfunc(*args):
        iargs = iter(args)
        pargs = [list(islice(iargs, pl)) for pl in pars_lens]
        values = [formatter(f(*a)) for f, a in zip(funcs, pargs)]
        return ' '.join(values)
    return retfunc, tuple(chain.from_iterable(pars))


def variable_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    if isinstance_m(self.parent, ['VarTuple']):
        if isinstance_m(self.parent.parameter, ['VarReference']):
            vars_ = self.parent.variables
            index = next(i for i, v in enumerate(vars_) if v == self)
            func, pars = checkref(self.parent.parameter)
            retfunc = (lambda *args: func(*args)[index], pars)
        else:
            assert isinstance_m(self.parent.parameter, ['Tuple'])
            retfunc = checkref(self.parameter)
    else:
        retfunc = checkref(self.parameter)
    logging.debug('variable_func:'+repr(self)+repr(retfunc))
    return retfunc


def number_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    if self.need_input:
        url = self.url
        filename = self.filename

        def number_loader():
            data = load_value(url, filename)
            if not isinstance(data, dict):
                raise TypeError('input must be of type '+str(dict))
            return load_object(data)
        retfunc = (number_loader, tuple())
        self.need_input = False
    else:
        pars = (self.inp_value, self.inp_units)
        retfunc = (lambda: typemap['Number'](*pars), tuple())
    logging.debug('number_func:'+repr(self)+repr(retfunc))
    return retfunc


def plain_type_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    if self.need_input:
        url = self.url
        filename = self.filename
        type_ = self.type_

        def plain_type_loader():
            data = load_value(url, filename)
            if not isinstance(data, type_):
                raise TypeError('input must be of type '+str(type_))
            return data
        retfunc = (plain_type_loader, tuple())
        self.need_input = False
    else:
        val = self.value  # evaluate here to allow serialization with dill
        retfunc = (lambda: val, tuple())
    logging.debug('plain_type_func:'+repr(self)+repr(retfunc))
    return retfunc


def series_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    if self.need_input:
        url = self.url
        filename = self.filename

        def series_loader():
            data = load_value(url, filename)
            if not isinstance(data, dict):
                raise TypeError('input must be of type '+str(dict))
            return load_object(data)
        retfunc = (series_loader, tuple())
        self.need_input = False
    else:
        name = self.name
        if self.type_.datatype is int:
            elems = (numpy.nan if not isinstance(e, int) else e for e in self.elements)
            data = [typemap['Number'](e, self.inp_units) for e in elems]
            retfunc = (lambda: typemap['Series'](data=data, name=name), tuple())
        elif self.type_.datatype is float:
            data = [numpy.nan if not isinstance(e, (int, float)) else e for e in self.elements]
            units = self.inp_units if self.inp_units else 'dimensionless'
            dtype = pint_pandas.PintType(units)
            retfunc = (lambda: typemap['Series'](data=data, dtype=dtype,
                       name=name), tuple())
        else:
            tups = [checkref(e) for e in self.elements]
            funs = [t[0] for t in tups]
            pars = [t[1] for t in tups]
            lens = [len(p) for p in pars]

            def get_series_val(*args):
                iargs = iter(args)
                fargs = [list(islice(iargs, pl)) for pl in lens]
                return typemap['Series']([f(*a) for f, a in zip(funs, fargs)], name=name)
            retfunc = (get_series_val, tuple(chain.from_iterable(pars)))
    return retfunc


def table_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    if self.need_input:
        url = self.url
        filename = self.filename

        def table_loader():
            data = load_value(url, filename)
            if not isinstance(data, dict):
                raise TypeError('input must be of type '+str(dict))
            return load_object(data)
        retfunc = (table_loader, tuple())
        self.need_input = False
    else:
        funcs = [c.func[0] for c in self.columns]
        pars = [c.func[1] for c in self.columns]
        pars_lens = [len(p) for p in pars]

        def get_table_val(*args):
            iargs = iter(args)
            fargs = [list(islice(iargs, pl)) for pl in pars_lens]
            series = [f(*a) for f, a in zip(funcs, fargs)]
            return typemap['Table'](pandas.concat(series, axis=1))
        retfunc = (get_table_val, tuple(chain.from_iterable(pars)))
    return retfunc


def bool_str_array_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    if self.url or self.filename:
        url = self.url
        filename = self.filename
        return (lambda: load_object(load_value(url, filename)), tuple())
    data = get_array_aslist(self.elements)
    return (lambda: numpy.array(data), tuple())


def int_array_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    if self.url or self.filename:
        url = self.url
        filename = self.filename
        return (lambda: load_object(load_value(url, filename)), tuple())
    data = get_array_aslist(self.elements, self.inp_units)
    return (lambda: numpy.asarray(data, dtype='object'), tuple())


def float_array_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    if self.url or self.filename:
        url = self.url
        filename = self.filename
        return (lambda: load_object(load_value(url, filename)), tuple())
    data = get_array_aslist(self.elements)
    units = self.inp_units if self.inp_units else 'dimensionless'
    return (lambda: typemap['FloatArray'](data, dtype=pint_pandas.PintType(units)), tuple())


def get_property_reference_func(func, pars, accessor):
    """return a 2-tuple containing a function and a list of parameters"""
    if accessor.index is not None:
        acc_index = accessor.index

        def get_property_index(value):
            if isinstance(value, typemap['Table']):
                dfr = value.iloc[[acc_index]]
                return tuple(next(dfr.itertuples(index=False, name=None)))
            if isinstance(value, typemap['Series']):
                return value.values[acc_index]
            if isinstance(value, amml.AMMLObject):
                return value[acc_index]
            raise TypeError(f'invalid type {type(value)}')
        return (lambda *x: get_property_index(func(*x)), pars)
    acc_id = accessor.id
    return (lambda *x: func(*x)[acc_id], pars)


def property_reference_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    func, pars = checkref(self.varref)
    for accessor in self.accessors:
        func, pars = get_property_reference_func(func, pars, accessor)
    return func, pars


def iterable_property_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    func, pars = checkref(self.obj)
    if hasattr(self, 'name_') and self.name_:
        retfunc = (lambda *x: func(*x).name, pars)
    elif hasattr(self, 'columns') and self.columns:
        retfunc = (lambda *x: tuple(func(*x).columns.to_list()), pars)
    else:
        slice_ = self.slice
        start_ = self.start
        stop_ = self.stop
        step_ = self.step

        def get_sliced_value(value):
            return value[start_:stop_:step_] if slice_ else value

        if self.array:
            if issubclass(self.obj.type_.datatype, str):
                retfunc = (lambda *x: get_sliced_value(func(*x)).values.astype(str), pars)
            else:
                retfunc = (lambda *x: get_sliced_value(func(*x)).values, pars)
        else:
            retfunc = (settype(lambda *x: get_sliced_value(func(*x))), pars)
    return retfunc


def iterable_query_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    func, pars = checkref(self.obj)
    pars = list(pars)
    olens = len(pars)
    where = self.where
    where_all = self.where_all
    where_any = self.where_any
    column_names = self.columns
    if where:
        conds = [self.condition] if self.condition else self.conditions
        cfunc = [c.func[0] for c in conds]
        cpars = [c.func[1] for c in conds]
        clens = [len(p) for p in cpars]
        pars.extend(chain.from_iterable(cpars))
    else:
        cfunc = []
        clens = []

    def get_query_value(*args):
        iargs = iter(args)
        oargs = list(islice(iargs, olens))
        value = func(*oargs)
        if where:
            fargs = [list(islice(iargs, cl)) for cl in clens]
            fconds = [f(*a) for f, a in zip(cfunc, fargs)]
            if where_all:
                val = value[reduce(lambda x, y: x & y, fconds)]
            elif where_any:
                val = value[reduce(lambda x, y: x | y, fconds)]
            else:
                val = value[fconds[0]]
        else:
            val = value
        if column_names:
            val = val[column_names]
        return val.reset_index(drop=True)
    retfunc = (get_query_value, tuple(pars))
    return retfunc


def condition_in_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    qobj_ref = get_parent_of_type('IterableQuery', self).obj
    qobj_ref_func, cpars = checkref(qobj_ref)
    column = self.column

    def cfunc(*args):
        val = qobj_ref_func(*args)
        if isinstance(val, pandas.DataFrame):
            return val[column]
        assert val.name == column
        return val

    cpars_len = len(cpars)
    if self.parameter:
        pfunc, ppars = checkref(self.parameter)

        def retfunc1(*args):
            return cfunc(*args[:cpars_len]).isin(pfunc(*args[cpars_len:]))
        retfunc = (retfunc1, (*cpars, *ppars))
    else:
        pfunc = [p.func[0] for p in self.params]
        ppars = tuple(p.func[1] for p in self.params)
        ppars_lens = [len(p) for p in ppars]

        def retfunc2(*args):
            iargs = iter(args)
            cargs = list(islice(iargs, cpars_len))
            pargs = [list(islice(iargs, pl)) for pl in ppars_lens]
            return cfunc(*cargs).isin([f(*a) for f, a in zip(pfunc, pargs)])
        retfunc = (retfunc2, tuple(chain.from_iterable((cpars,)+ppars)))
    logging.debug(repr(self)+repr(retfunc))
    return retfunc


def condition_comparison_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    operator = binop_map[self.operator]
    qobj_ref = get_parent_of_type('IterableQuery', self).obj
    qobj_func, qobj_pars = checkref(qobj_ref)
    qobj_parlen = len(qobj_pars)
    if self.operand_right:
        opr_func, opr_pars = checkref(self.operand_right)
    if self.operand_left:
        opl_func, opl_pars = checkref(self.operand_left)
    left = self.column_left
    right = self.column_right
    if issubclass(qobj_ref.type_, typemap['Table']):
        if self.column_left and self.column_right:

            def retfunc1(*args):
                val = qobj_func(*args)
                return operator(val[left], val[right])
            retfunc = (retfunc1, qobj_pars)
        elif left and self.operand_right:

            def retfunc2(*args):
                val = qobj_func(*args[0:qobj_parlen])
                return operator(val[left], opr_func(*args[qobj_parlen:]))
            retfunc = (retfunc2, (*qobj_pars, *opr_pars))
        elif self.operand_left and right:

            def retfunc3(*args):
                val = qobj_func(*args[0:qobj_parlen])
                return operator(opl_func(*args[qobj_parlen:]), val[right])
            retfunc = (retfunc3, (*qobj_pars, *opr_pars))
    else:
        assert issubclass(qobj_ref.type_, typemap['Series'])
        if left and self.operand_right:

            def retfunc4(*args):
                val = qobj_func(*args[0:qobj_parlen])
                assert val.name == left
                return operator(val, opr_func(*args[qobj_parlen:]))
            retfunc = (retfunc4, (*qobj_pars, *opr_pars))
        elif self.operand_left and right:

            def retfunc5(*args):
                val = qobj_func(*args[0:qobj_parlen])
                assert val.name == right
                return operator(opl_func(*args[qobj_parlen:]), val)
            retfunc = (retfunc5, (*qobj_pars, *opl_pars))
    return retfunc


def range_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    tup = (self.start, self.stop, self.step)
    tups = [checkref(t) for t in tup]
    funcs = [t[0] for t in tups]
    pars = [t[1] for t in tups]
    lens = [len(t[1]) for t in tups]
    name = self.parent.name if hasattr(self.parent, 'name') else None

    def get_range_val(*args):
        iargs = iter(args)
        fargs = [list(islice(iargs, pl)) for pl in lens]
        cargs = [f(*a) for f, a in zip(funcs, fargs)]
        unit = cargs[0].units
        data = numpy.arange(cargs[0].magnitude, cargs[1].to(unit).magnitude,
                            cargs[2].to(unit).magnitude).tolist()
        # redundant code with series, instant executor
        if all(isinstance(e, int) for e in data):
            data = (typemap['Number'](e, unit) for e in data)
            dtype = 'object'
        elif all(isinstance(e, float) for e in data):
            dtype = pint_pandas.PintType(unit)
        return typemap['Series'](data=data, name=name, dtype=dtype)
    return (get_range_val, tuple(chain.from_iterable(pars)))


def map_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    name = self.parent.name if hasattr(self.parent, 'name') else None
    func_def = self.lambda_ if self.lambda_ else self.function
    if not isinstance_m(func_def, ['ObjectImport']):
        nfunc, npars = dummies_right(*checkref(func_def.expr))
        func_args = func_def.args
    else:
        nfunc = get_object_import(self.function)
        dummy = get_metamodel(self)['Dummy']
        dummy.name = None
        npars = (dummy,)*len(self.params)
        func_args = npars
    func_pars = [checkref(p) for p in self.params]
    pars_lens = sum(not isinstance_m(p, ['Dummy']) for p in npars)
    non_dummy_pars = npars[:pars_lens]
    dummy_pars = []
    dummy_funcs = []
    for par in npars[pars_lens:]:
        ind = next(i for i, a in enumerate(func_args) if a.name == par.name)
        dummy_pars.append(func_pars[ind][1])
        dummy_funcs.append(func_pars[ind][0])
    dummy_pars_lens = [len(p) for p in dummy_pars]
    all_pars = [non_dummy_pars] + dummy_pars

    def get_map_val(*args):
        iargs = iter(args)
        pargs = list(islice(iargs, pars_lens))
        dargs = [list(islice(iargs, pl)) for pl in dummy_pars_lens]
        iterables = [f(*a) for f, a in zip(dummy_funcs, dargs)]
        return typemap['Series'](map(partial(nfunc, *pargs), *iterables), name=name)
    return (get_map_val, tuple(chain.from_iterable(all_pars)))


def filter_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    name = self.parent.name if hasattr(self.parent, 'name') else None
    func_def = self.lambda_ if self.lambda_ else self.function
    if not isinstance_m(func_def, ['ObjectImport']):
        nfunc, npars = dummies_right(*checkref(func_def.expr))
    else:
        nfunc = get_object_import(self.function)
        npars = tuple()
    pfun, pars = checkref(self.parameter)
    plen = sum(not isinstance_m(p, ['Dummy']) for p in npars)
    all_pars = npars[:plen] + pars
    retfunc = (lambda *x: typemap['Series'](filter(partial(nfunc, *x[:plen]),
               pfun(*x[plen:])), name=name), all_pars)
    return retfunc


def reduce_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    func_def = self.lambda_ if self.lambda_ else self.function
    pfun, pars = checkref(self.parameter)
    if not isinstance_m(func_def, ['ObjectImport']):
        nfunc, npars = dummies_right(*checkref(func_def.expr))
        plen = sum(not isinstance_m(p, ['Dummy']) for p in npars)
        npars_u = npars[:plen] + tuple(func_def.args)
        mapping = [next(i for i, a in enumerate(npars_u) if a == p) for p in npars]

        def nfunc_u(*args):
            return partial(lambda *x: nfunc(*[x[i] for i in mapping]), *args)
        apars = npars[:plen] + pars
        retfunc = (lambda *x: reduce(nfunc_u(*x[:plen]), pfun(*x[plen:])), apars)
    else:
        nfunc = get_object_import(self.function)
        retfunc = (lambda *x: reduce(nfunc, pfun(*x)), pars)
    return retfunc


def func_reduce_func(self, func):
    """return a 2-tuple containing a function and a list of parameters"""
    if self.parameter:
        pfun, pars = checkref(self.parameter)
        retfunc = (lambda *x: func(pfun(*x)), pars)
    else:
        tuples = [checkref(p) for p in self.params]
        funcs = [t[0] for t in tuples]
        pars = [t[1] for t in tuples]
        pars_lens = [len(p) for p in pars]

        def get_func_reduce_val(*args):
            iargs = iter(args)
            fargs = [list(islice(iargs, pl)) for pl in pars_lens]
            return func([f(*a) for f, a in zip(funcs, fargs)])
        retfunc = (get_func_reduce_val, tuple(chain.from_iterable(pars)))
    return retfunc


def in_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    if self.parameter:
        func_x, pars_x = checkref(self.element)
        func_y, pars_y = checkref(self.parameter)
        len_x = len(pars_x)
        retfunc = (lambda *args: func_x(*args[:len_x]) in func_y(*args[len_x:]).values,
                   (*pars_x, *pars_y))
    else:
        elem_func, elem_pars = checkref(self.element)
        elem_pars_len = len(elem_pars)
        func_pars = [checkref(p) for p in self.params]
        funcs = [t[0] for t in func_pars]
        pars = tuple(t[1] for t in func_pars)
        pars_lens = [len(p) for p in pars]

        def get_in_val(*args):
            iargs = iter(args)
            eargs = list(islice(iargs, elem_pars_len))
            pargs = [list(islice(iargs, pl)) for pl in pars_lens]
            return elem_func(*eargs) in [f(*a) for f, a in zip(funcs, pargs)]
        retfunc = (get_in_val, tuple(chain.from_iterable(elem_pars+pars)))
    return retfunc


def binop_func(func, ops, from_the_left=True):
    """binary operator as a 2-tuple with a function and parameters"""
    if len(ops) == 0:
        retval = func
    else:
        operator, operand = ops.pop(0) if from_the_left else ops.pop(-1)
        left_func, left_pars = func if from_the_left else checkref(operand)
        right_func, right_pars = checkref(operand) if from_the_left else func
        pars = left_pars + right_pars
        operator = binop_map[operator]
        left_pars_length = len(left_pars)
        retval = binop_func(
            (
                lambda *args: operator(left_func(*args[0:left_pars_length]),
                                       right_func(*args[left_pars_length:])),
                pars
            ),
            ops, from_the_left=from_the_left)
    return retval


def expression_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    func = checkref(self.operands[0])
    ops = list(zip(self.operators, self.operands[1:]))
    retfunc = binop_func(func, ops)
    logging.debug('expression_func:'+repr(self)+repr(retfunc))
    return retfunc


def term_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    func = checkref(self.operands[0])
    ops = list(zip(self.operators, self.operands[1:]))
    retfunc = binop_func(func, ops)
    logging.debug('term_func:'+repr(self)+repr(retfunc))
    return retfunc


def factor_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    func = checkref(self.operands[-1])
    operators = ['**']*(len(self.operands)-1)
    ops = list(zip(operators, self.operands[:-1]))
    retfunc = binop_func(func, ops, from_the_left=False)
    logging.debug('factor_func:'+repr(self)+repr(retfunc))
    return retfunc


def power_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    func, pars = checkref(self.operand)
    if self.sign == '-':
        retfunc = (lambda *args: -func(*args), pars)
    else:
        retfunc = (func, pars)
    logging.debug('power_func:'+repr(self)+repr(retfunc))
    return retfunc


def operand_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    return self.operand.func


def binary_operation_func(self, operator):
    """return a 2-tuple containing a function and a list of parameters"""
    func = checkref(self.operands[0])
    ops = list(zip([operator]*len(self.operands[1:]), self.operands[1:]))
    retfunc = binop_func(func, ops)
    logging.debug(repr(self)+repr(retfunc))
    return retfunc


def not_func(self, operator):
    """return a 2-tuple containing a function and a list of parameters"""
    retfunc = checkref(self.operand)
    if self.not_:
        func, pars = retfunc
        retfunc = (lambda *args: operator(func(*args)), pars)
    logging.debug(repr(self)+repr(retfunc))
    return retfunc


def comparison_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    func = checkref(self.left)
    ops = [(self.operator, self.right)]
    new_func, new_pars = binop_func(func, ops)
    type_ = self.type_
    return (lambda *args: type_(new_func(*args)), new_pars)


def if_expression_func(self):
    """
    Naive implementation: condition and the active branch are executed serially
    on the same resources
    """
    expr_func, expr_pars = checkref(self.expr)
    expr_pars_len = len(expr_pars)
    true_b, true_b_pars = checkref(self.true_)
    true_b_pars_len = expr_pars_len + len(true_b_pars)
    false_b, false_b_pars = checkref(self.false_)

    def iffunc(*args):
        if expr_func(*args[:expr_pars_len]):
            retval = true_b(*args[expr_pars_len:true_b_pars_len])
        else:
            retval = false_b(*args[true_b_pars_len:])
        return retval
    retfunc = (iffunc, (*expr_pars, *true_b_pars, *false_b_pars))
    logging.debug(repr(self)+repr(retfunc))
    return retfunc


def function_call_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    if isinstance_m(self.function, ['ObjectImport']):
        obj = get_object_import(self.function)
        assert callable(obj)
        func_pars = [checkref(p) for p in self.params]
        pars = [pars for func, pars in func_pars]
        funcs = [func for func, pars in func_pars]
        lens = list(map(len, pars))
        pars_flat = tuple(p for par in pars for p in par)

        def fcall_func(*args):
            iter_args = iter(args)
            fcall_args = [list(islice(iter_args, i)) for i in lens]
            return obj(*[f(*a) for f, a in zip(funcs, fcall_args)])
        retfunc = (settype(fcall_func), pars_flat)
    elif isinstance_m(self.function, ['FunctionDefinition']):
        if not get_parent_of_type('FunctionDefinition', self):
            retfunc = checkref(self.expr)
        else:
            # function call in a function definition, e.g. f(x) = 2*g(x)
            retfunc = None
    logging.debug(repr(self)+repr(retfunc))
    return retfunc


def object_import_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    obj = get_object_import(self)
    assert not callable(obj)
    return (settype(lambda: obj), tuple())


def tuple_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    tuple_pars = [checkref(p) for p in self.params]
    tpars = [pars for func, pars in tuple_pars]
    funcs = [func for func, pars in tuple_pars]
    tlens = [len(pars) for func, pars in tuple_pars]

    def get_tuple_val(*args):
        iter_args = iter(args)
        targs = [list(islice(iter_args, t)) for t in tlens]
        return tuple(f(*a) for f, a in zip(funcs, targs))
    return (get_tuple_val, tuple(chain.from_iterable(tpars)))


def amml_structure_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    if self.filename or self.url:
        url = self.url
        filename = self.filename
        suffix = pathlib.Path(filename).suffix
        if filename and suffix not in ['.yml', '.yaml', '.json']:
            return (lambda: amml.AMMLStructure.from_ase_file(filename), tuple())
        return (lambda: load_object(load_value(url, filename)), tuple())
    func, pars = checkref(self.tab)
    name = self.name
    return (lambda *x: amml.AMMLStructure(func(*x), name), pars)


def amml_calculator_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    if self.filename or self.url:
        url = self.url
        filename = self.filename
        return (lambda: load_object(load_value(url, filename)), tuple())
    name = self.name
    pinning = self.pinning
    version = self.version
    if self.parameters is None:
        return (lambda: amml.Calculator(name, None, pinning, version), tuple())
    func, pars = checkref(self.parameters)
    return (lambda *x: amml.Calculator(name, func(*x), pinning, version), pars)


def amml_property_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    cfunc, cpars = checkref(self.calc)
    sfunc, spars = checkref(self.struct)
    props = self.names
    cpl = len(cpars)
    spl = len(spars)
    constr_tuples = [checkref(c) for c in self.constrs]
    constr_func = [f for f, p in constr_tuples]
    constr_pars = [p for f, p in constr_tuples]
    constr_pars_lens = [len(p) for p in constr_pars]
    pars = (cpars, spars, *constr_pars)
    task = self.task

    def get_property_val(*args):
        iargs = iter(args)
        calc_args = list(islice(iargs, cpl))
        struct_args = list(islice(iargs, spl))
        constr_args = [list(islice(iargs, pl)) for pl in constr_pars_lens]
        constrs = [f(*x) for f, x in zip(constr_func, constr_args)]
        return amml.Property(props, cfunc(*calc_args), sfunc(*struct_args),
                             constrs=constrs, task=task)
    return (get_property_val, tuple(chain.from_iterable(pars)))


def amml_constraint_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    name = self.name
    fixed_func, fixed_pars = checkref(self.fixed)
    if self.direction is None:
        return (lambda *x: amml.Constraint(name, fixed=fixed_func(*x)), fixed_pars)
    fpl = len(fixed_pars)
    direc_func, direc_pars = checkref(self.direction)
    pars = (*fixed_pars, *direc_pars)
    return (lambda *x: amml.Constraint(name, fixed=fixed_func(*x[:fpl]),
                                       direction=direc_func(*x[fpl:])), pars)


def chem_reaction_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    species = [checkref(term.species) for term in self.educts+self.products]
    func_spec = [t[0] for t in species]
    pars_spec = tuple(t[1] for t in species)
    pars_lens = [len(p) for p in pars_spec]
    coeffs = []
    for term in self.educts:
        coeffs.append(-term.coefficient)
    for term in self.products:
        coeffs.append(term.coefficient)

    if self.props is None:
        props_avail = False
        func_props = None
        pars_all = pars_spec
    else:
        props_avail = True
        func_props, pars_props = checkref(self.props)
        pars_all = pars_spec + pars_props

    def get_chem_reaction(*args):
        iargs = iter(args)
        pargs = [list(islice(iargs, pl)) for pl in pars_lens]
        specs = [f(*a) for f, a in zip(func_spec, pargs)]
        terms = [{'coefficient': c, 'species': s} for c, s in zip(coeffs, specs)]
        props = func_props(*list(iargs)) if props_avail else None
        return chemistry.ChemReaction(terms, props)
    return get_chem_reaction, tuple(chain.from_iterable(pars_all))


def chem_species_func(self):
    """return a 2-tuple containing a function and a list of parameters"""
    name = self.name
    if self.props is None:
        return (lambda: chemistry.ChemSpecies(name), tuple())
    func, pars = checkref(self.props)
    return (lambda *x: chemistry.ChemSpecies(name, func(*x)), pars)


def add_func_properties(metamodel):
    """Add class properties using monkey style patching"""
    mapping_dict = {
        'Null': lambda _: (lambda: None, tuple()),
        'Print': print_func,
        'VarReference': reference_func,
        'Variable': variable_func,
        'String': plain_type_func,
        'Bool': plain_type_func,
        'Number': number_func,
        'Power': power_func,
        'Factor': factor_func,
        'Term': term_func,
        'Expression': expression_func,
        'Operand': operand_func,
        'BooleanOperand': operand_func,
        'And': partial(binary_operation_func, operator='and'),
        'Or': partial(binary_operation_func, operator='or'),
        'Not': partial(not_func, operator=not_),
        'Comparison': comparison_func,
        'IfFunction': if_expression_func,
        'IfExpression': if_expression_func,
        'FunctionCall': function_call_func,
        'Lambda': lambda x: checkref(x.expr),
        'FunctionDefinition': lambda x: checkref(x.expr),
        'ObjectImport': object_import_func,
        'Tuple': tuple_func,
        'Series': series_func,
        'Table': table_func,
        'BoolArray': bool_str_array_func,
        'StrArray': bool_str_array_func,
        'IntArray': int_array_func,
        'FloatArray': float_array_func,
        'PropertyReference': property_reference_func,
        'IterableProperty': iterable_property_func,
        'IterableQuery': iterable_query_func,
        'ConditionIn': condition_in_func,
        'ConditionComparison': condition_comparison_func,
        'ConditionOr': partial(binary_operation_func, operator='or_'),
        'ConditionAnd': partial(binary_operation_func, operator='and_'),
        'ConditionNot': partial(not_func, operator=invert),
        'Range': range_func,
        'In': in_func,
        'Any': partial(func_reduce_func, func=any),
        'All': partial(func_reduce_func, func=all),
        'Sum': partial(func_reduce_func, func=sum),
        'Map': map_func,
        'Filter': filter_func,
        'Reduce': reduce_func,
        'AMMLStructure': amml_structure_func,
        'AMMLCalculator': amml_calculator_func,
        'AMMLProperty': amml_property_func,
        'AMMLConstraint': amml_constraint_func,
        'ChemReaction': chem_reaction_func,
        'ChemSpecies': chem_species_func
    }
    for key, function_ptr in mapping_dict.items():
        metamodel[key].func = cached_property(function_ptr)
        metamodel[key].func.__set_name__(metamodel[key], 'func')


@checktype
def func_value(self):
    """evaluate a python function object"""
    logging.debug('func_value:%s', repr(self))
    func, pars = self.func
    return func(*[p.value for p in pars])


def add_deferred_value_properties(metamodel):
    """Add class properties using monkey style patching"""
    mapping_dict = {
        # reused from instant executor
        'Program': program_value,
        'VarReference': func_value,
        'String': plain_type_value,
        'Bool': plain_type_value,
        # deferred evaluation via a python function object
        'Null': func_value,
        'Number': func_value,
        'Print': func_value,
        'Variable': func_value,
        'Power': func_value,
        'Factor': func_value,
        'Term': func_value,
        'Expression': func_value,
        'Operand': func_value,
        'BooleanOperand': func_value,
        'And': func_value,
        'Or': func_value,
        'Not': func_value,
        'Comparison': func_value,
        'IfFunction': func_value,
        'IfExpression': func_value,
        'FunctionCall': func_value,
        'ObjectImport': func_value,
        'Tuple': func_value,
        'Series': func_value,
        'Table': func_value,
        'BoolArray': func_value,
        'StrArray': func_value,
        'IntArray': func_value,
        'FloatArray': func_value,
        'IterableProperty': func_value,
        'IterableQuery': func_value,
        'ConditionIn': func_value,
        'ConditionComparison': func_value,
        'ConditionOr': func_value,
        'ConditionAnd': func_value,
        'ConditionNot': func_value,
        'Range': func_value,
        'In': func_value,
        'Any': func_value,
        'All': func_value,
        'Sum': func_value,
        'Map': func_value,
        'Filter': func_value,
        'Reduce': func_value,
        'AMMLStructure': func_value,
        'AMMLCalculator': func_value,
        'AMMLProperty': func_value,
        'AMMLConstraint': func_value,
        'ChemReaction': func_value,
        'ChemSpecies': func_value
    }
    for key, func in mapping_dict.items():
        metamodel[key].value = cached_property(textxerror_wrap(func))
        metamodel[key].value.__set_name__(metamodel[key], 'value')
