"""manage sessions for dynamic model processing / incremental development"""
import logging
import warnings
from dateutil import tz
from textx import metamodel_from_file, metamodel_from_str, textx_isinstance, get_children
from virtmat.language.metamodel.properties import add_properties
from virtmat.language.metamodel.processors import add_processors
from virtmat.language.utilities.textx import GrammarString, isinstance_m
from virtmat.language.utilities.errors import error_handler
from virtmat.middleware.engine.wfengine import WFEngine


class Session:
    """session for dynamic model processing"""
    model = None
    wfe = None

    def __init__(self, lpad, uuid=None, grammar_str=None, grammar_path=None,
                 model_str=None, model_path=None, create_new=False, autorun=False,
                 on_demand=False, async_run=False, textx_debug=False, **wfe_kwargs):
        self.lpad = lpad
        self.uuid = uuid
        self.async_run = async_run
        self.textx_debug = textx_debug
        tx_kwargs = {'auto_init_attributes': False, 'debug': textx_debug}
        if self.uuid is not None:
            if grammar_str or grammar_path:
                msg = 'provided grammar ignored in favor of grammar from provided uuid'
                logging.warning(msg)
                warnings.warn(msg, UserWarning)
            wf_ids = self.lpad.get_wf_ids({'metadata.uuid': self.uuid})
            if len(wf_ids) == 0:
                msg = 'no model with provided uuid could be found'
                logging.error(msg)
                raise RuntimeError(msg)
            assert len(wf_ids) == 1
            workflow = self.lpad.get_wf_by_fw_id(wf_ids[0])  # todo: use less expensive query
            self.grammar_str = workflow.metadata['grammar_str']
            self.metamodel = metamodel_from_str(self.grammar_str, **tx_kwargs)
        else:
            if grammar_path:
                self.metamodel = metamodel_from_file(grammar_path, **tx_kwargs)
                self.grammar_str = GrammarString(grammar_path).string
            elif grammar_str:
                self.metamodel = metamodel_from_str(grammar_str, **tx_kwargs)
                self.grammar_str = grammar_str
            else:
                msg = 'grammar must be provided if uuid not provided'
                logging.error(msg)
                raise ValueError(msg)
        self.modify_metamodel()
        if create_new or (self.async_run and self.uuid is None):
            # create a root node and complete it
            self.autorun = True
            self.on_demand = False
            self.get_model('print(true)')
        self.autorun = autorun
        self.on_demand = on_demand
        if self.autorun and self.async_run:
            assert self.uuid is not None
            wf_query = {'metadata.uuid': self.uuid}
            self.wfe = WFEngine(lpad, wf_query=wf_query, **wfe_kwargs)
            if self.on_demand:
                self.wfe.nodes_torun = []
            self.start_runner()
        self.get_model(model_str, model_path)

    def start_runner(self):
        """a helper function to activate evaluation"""
        if self.async_run:
            self.wfe.start()
            self.autorun = False
        else:
            self.autorun = True

    def stop_runner(self):
        """a helper function to deactivate evaluation"""
        if self.async_run:
            self.wfe.stop(join=True)
        else:
            self.autorun = False

    def modify_metamodel(self):
        """add class properties, object and model processors to metamodel"""
        add_properties(self.metamodel, deferred_mode=True, workflow_mode=True)
        add_processors(self.metamodel, wflow_processors=True)

    def get_model_nodes(self):
        """return the list of nodes in the model"""
        wf_query = {'metadata.uuid': self.uuid}
        fw_query = {'spec._source_code': {'$exists': True}}
        fw_ids = self.lpad.get_fw_ids_in_wfs(wf_query, fw_query)
        return [self.lpad.get_fw_by_id(i) for i in fw_ids]

    def get_model_str(self, model_str=''):
        """updates the current model string with model_str"""
        if self.uuid is not None:
            fws = self.get_model_nodes()
            lines = [line for fwk in fws for line in fwk.spec['_source_code']]
            pers_model_str = '\n'.join(list(set(lines))) + '\n'
        else:
            pers_model_str = ''
        return pers_model_str + model_str

    def get_model_history(self):
        """return a formatted node history with some node attributes"""
        fws = self.get_model_nodes()
        # mongodb stores all times in UTC so we need to convert to local
        utc_zone = tz.tzutc()
        our_zone = tz.tzlocal()
        lines = []
        for fwk in fws:
            loc_time = fwk.updated_on.replace(tzinfo=utc_zone).astimezone(our_zone)
            iso_time = loc_time.isoformat(timespec='seconds')
            for line in fwk.spec['_source_code']:
                lines.append(' '.join((f'{fwk.state:10}', iso_time, line)))
        return '\n'.join(lines)

    def get_model(self, model_str=None, model_path=None):
        """returns the model object; if model_str/model_path supplied creates a
           new textx model instance"""
        if model_str or model_path or (self.model is None and self.uuid is not None):
            if model_str:
                tx_get_model = self.metamodel.model_from_str
                model_src = self.get_model_str(model_str)
                source_code = model_src
            elif model_path:
                with open(model_path, 'r', encoding='utf-8') as inp:
                    model_str = inp.read()
                if self.uuid is not None:
                    tx_get_model = self.metamodel.model_from_str
                    model_src = self.get_model_str(model_str)
                    source_code = model_src
                else:
                    tx_get_model = self.metamodel.model_from_file
                    model_src = model_path
                    source_code = model_str
            else:
                tx_get_model = self.metamodel.model_from_str
                model_src = self.get_model_str()
                source_code = model_src

            model_instance = {'lpad': self.lpad, 'uuid': self.uuid}
            self.model = tx_get_model(model_src, deferred_mode=True,
                                      model_instance=model_instance,
                                      source_code=source_code,
                                      grammar_str=self.grammar_str,
                                      autorun=self.autorun,
                                      on_demand=self.on_demand,
                                      debug=self.textx_debug)
            self.uuid = getattr(self.model, 'uuid', None)
            if not textx_isinstance(self.model, self.metamodel['Program']):
                self.model = None
            elif self.wfe and self.on_demand:
                new_nodes = set(self.wfe.nodes_torun+self.model.fw_ids_torun)
                self.wfe.nodes_torun = list(new_nodes)

            if self.model:
                name_list = []
                classes = ['Variable', 'FunctionDefinition', 'ObjectImport']
                names = get_children(lambda x: isinstance_m(x, classes),
                                     self.model)
                for name_ in names:
                    name_list.append(name_.name)
                self.model.name_list = name_list

        return self.model


class SessionManager():
    """session manager for basic interactive work using a text terminal"""

    def __init__(self, lpad, **kwargs):
        self.lpad = lpad
        self.kwargs = dict(kwargs)
        del self.kwargs['uuid']
        del self.kwargs['grammar_path']
        self.session = Session(lpad, create_new=True, **kwargs)

    @error_handler
    def get_model_value(self, *args, **kwargs):
        """wrapped and evaluated version of get_model() of the Session class"""
        return getattr(self.session.get_model(*args, **kwargs), 'value', '')

    def main_loop(self):
        """this is the main loop of the interactive session"""
        while True:
            try:
                input_str = input('Input > ').strip()
            except EOFError:
                print('\nExiting')
                self.session.stop_runner()
                break
            except KeyboardInterrupt:
                print('\n')
                continue
            if input_str == '%stop':
                self.session.stop_runner()
                continue
            if input_str == '%start':
                self.session.start_runner()
                continue
            if '%uuid' in input_str:
                new_uuid = input_str.strip().split()[1]
                self.session = Session(self.lpad, uuid=new_uuid, **self.kwargs)
                continue
            if input_str in ('%exit', '%bye', '%close', '%quit'):
                print('Exiting')
                self.session.stop_runner()
                break
            if input_str in ('%hist', '%history'):
                print(self.session.get_model_history())
                continue
            if '"""' in input_str:
                print('multi-line comments not supported')
                continue
            if not any(word in input_str for word in ('print', 'use', 'to', '=', '#')):
                # ipython-like print statement
                input_str = f'print({input_str})'
            output = self.get_model_value(model_str=input_str)
            if output:
                print('Output >', output)
