# pylint: disable = protected-access, fixme
"""
Language interpreter for distributed/remote evaluation using workflow
management systems and batch systems.

Variables store outputs of function tasks, therefore a variable is in the root
of every evaluation (-> mapped to a task in the workflow). Grouping tasks to
nodes can be done either via simple mapping line of code -> node, or based on
graph analysis.
"""
import logging
import base64
import uuid
from functools import cached_property
import dill
import pandas
from textx import get_children_of_type, get_model
from textx import get_location
from textx.exceptions import TextXError
from fireworks import Firework, Workflow
from virtmat.language.utilities.firetasks import FunctionTask, ExportDataTask, ScatterTask
from virtmat.language.utilities.fw_tools import run_fireworks, get_ancestors
from virtmat.language.utilities.fw_tools import get_nodes_providing, get_parent_nodes
from virtmat.language.utilities.serializable import get_serializable, get_basetyped
from virtmat.language.utilities.textx import isinstance_m
from virtmat.language.utilities.formatters import formatter
from virtmat.language.utilities.errors import print_stderr, textxerror_wrap
from virtmat.language.utilities.types import checktype
from .instant_executor import program_value, number_value, plain_type_value

# LOGLEVEL = logging.INFO
LOGLEVEL = logging.DEBUG
logging.basicConfig(level=LOGLEVEL)


class NonCompletedException(Exception):
    """this exception should be raised if a variable is not yet evaluated"""


def get_input_name(par):
    """return the input name and input value of a named object"""
    inp_value = None
    if isinstance_m(par, ['ObjectImport']):
        inp_name = par.name
        inp_value = get_serializable(par.value)
    elif isinstance_m(par, ['Variable']):
        inp_name = par.name
    else:
        msg = 'parameter must be variable or imported object'
        logging.error('get_input_name: error: %s', msg)
        raise RuntimeError(msg)
    return inp_name, inp_value


def get_fstr(func):
    """Return a pickle-serialized function as a Python3 string"""
    try:
        f_str = base64.b64encode(dill.dumps(func)).decode('utf-8')
    except dill.PicklingError:
        logging.error('get_fstr: pickling error')
        raise
    return f_str


def get_fws(self):
    """Create one or more Firework objects for one Variable object"""
    logging.debug('get_fws: processing %s', repr(self.func))
    assert isinstance(self.func, tuple) and len(self.func) == 2
    func, pars = self.func
    logging.debug('get_fws: func: %s pars: %s', self.name, pars)
    self.__fw_name = str(uuid.uuid4())
    spec_inp = {}
    inputs = []
    for par in pars:
        inp_name, inp_value = get_input_name(par)
        inputs.append(inp_name)
        if inp_value is not None:
            spec_inp[inp_name] = inp_value
    spec_rsc = {'_source_code': getattr(self, 'source_code', None)}
    if self.resources is None:
        spec_rsc['_category'] = 'interactive'
    else:
        spec_rsc['_category'] = self.resources.category
        if spec_rsc['_category'] == 'batch':
            spec_rsc['_queueadapter'] = self.resources.qadapter
            spec_rsc['_queueadapter']['job_name'] = self.__fw_name
    if hasattr(self, 'nchunks'):
        fws = self._get_fws_parallel(inputs, spec_rsc)
    else:
        tsk = FunctionTask(func=get_fstr(func), inputs=inputs, outputs=[self.name])
        fws = [Firework([tsk], spec={**spec_rsc, **spec_inp}, name=self.__fw_name)]
    return fws


def _get_fws_parallel(self, inputs, spec_rsc):
    """parallel Map, Filter and Reduce as parameters of Variable (self)"""
    logging.debug('get_fws_in_parallel: func: %s', self.name)
    if isinstance_m(self.parameter, ('Map',)):
        split = [p.ref.name for p in self.parameter.params]
    else:
        split = [self.parameter.parameter.ref.name]
    chunk_ids = [f'{self.name}_chunk_{c}' for c in range(self.nchunks)]
    tsk = ScatterTask(func=get_fstr(self.func[0]), inputs=inputs,
                      chunk_ids=chunk_ids, split=split, spec=spec_rsc)
    spec = {'_category': 'interactive', '_source_code': spec_rsc['_source_code']}
    fw1 = Firework([tsk], spec=spec, name=str(uuid.uuid4()))
    if isinstance_m(self.parameter, ('Map', 'Filter')):
        fstr = get_fstr(lambda *x: get_serializable(pandas.concat(x)))
        tsks = [FunctionTask(func=fstr, inputs=chunk_ids, outputs=[self.name])]
    else:
        assert isinstance_m(self.parameter, ('Reduce',))
        outp_reduce = self.name + '_for_reduce'
        fstr = get_fstr(lambda *x: [get_serializable(elem) for elem in x])
        tsk1 = FunctionTask(func=fstr, inputs=chunk_ids, outputs=[outp_reduce])
        inps3 = [outp_reduce if i == self.name else i for i in inputs]
        tsk2 = FunctionTask(func=get_fstr(self.func[0]), inputs=inps3,
                            outputs=[self.name])
        tsks = [tsk1, tsk2]
    fw2 = Firework(tsks, spec=spec, name=self.__fw_name)
    return [fw1, fw2]


def get_fw_object_to(self):
    """Create a single Firework object for a ObjectTo object"""
    spec = {'_source_code': getattr(self, 'source_code', None)}
    tsk = ExportDataTask(varname=self.ref.name, filename=self.filename, url=self.url)
    self.__fw_name = str(uuid.uuid4())
    return Firework([tsk], spec=spec, name=self.__fw_name)


def get_wf(self):
    """create a workflow object from a model"""
    assert self._tx_model_params['model_instance']['uuid'] is None
    fws = []
    for obj in get_children_of_type('Variable', self):
        fws.extend(obj.fireworks)
    data_nodes = [o.firework for o in get_children_of_type('ObjectTo', self)]
    outs = {}
    for fwk in fws:
        lst = []
        for task in fwk.tasks:
            if 'outputs' in task:
                lst += task['outputs']
            elif 'chunk_ids' in task:
                lst += task['chunk_ids']
        outs[str(fwk.fw_id)] = lst
    inps = {str(f.fw_id): [i for t in f.tasks for i in t['inputs']] for f in fws}
    expi = {str(f.fw_id): [t['varname'] for t in f.tasks] for f in data_nodes}
    expo = {str(f.fw_id): [] for f in data_nodes}
    inps.update(expi)
    outs.update(expo)
    fws.extend(data_nodes)
    links = {}
    # naive implementation of the join
    for ofw in fws:
        oid = ofw.fw_id
        if len(outs[str(oid)]) > 0:
            links[str(oid)] = []
            for ifw in fws:
                for inp in inps[str(ifw.fw_id)]:
                    if inp in outs[str(oid)]:
                        links[str(oid)].append(ifw.fw_id)
    # find all root nodes
    root_fws = []
    for firework in fws:
        if all('inputs' in t and len(t['inputs']) == 0 for t in firework.tasks):
            root_fws.append(firework.fw_id)
    # create a new empty root node
    f_str = base64.b64encode(dill.dumps(lambda: None)).decode('utf-8')
    root_node = Firework([FunctionTask(func=f_str, inputs=[], outputs=[])],
                         name='_fw_root_node')
    # link all root nodes to the new root node
    links[str(root_node.fw_id)] = root_fws
    fws.append(root_node)
    grammar_str = self._tx_model_params.get('grammar_str')
    metadata = {'uuid': self.uuid, 'grammar_str': grammar_str}
    return Workflow(fws, links_dict=links, metadata=metadata)


@textxerror_wrap
@checktype
def variable_value(self):
    """obtain an output value from database: asynchronously and non-blocking"""
    model = get_model(self)
    fw_q = {'name': self.__fw_name}
    firework = model.lpad.get_fw_by_id(model.lpad.get_fw_ids(fw_q)[0])
    if firework.state == 'COMPLETED':
        action = firework.launches[-1].action
        ret_value = get_basetyped(action.update_spec[self.name])
    elif firework.state == 'FIZZLED':
        action = firework.launches[-1].action
        if '_exception' in action.stored_data:
            logging.error('variable_value:%s evaluation error', repr(self))
            exception_dct = action.stored_data['_exception']
            if exception_dct.get('_details') is None:
                print_stderr(exception_dct['_stacktrace'])
                raise RuntimeError('no _details in _exception found')
            pkl = exception_dct['_details']['pkl']
            raise dill.loads(base64.b64decode(pkl.encode()))
        raise RuntimeError('state FIZZLED but no exception found')
    elif firework.state == 'WAITING':
        parents = get_ancestors(model.lpad, firework.fw_id)
        fw_query = {'fw_id': {'$in': parents}, 'state': 'FIZZLED'}
        fizzled = model.lpad.get_fw_ids(fw_query)
        if fizzled:
            msg = 'Evaluation not possible due to failed ancestors: '
            var_names = []
            vars_ = get_children_of_type('Variable', model)
            for fw_id in fizzled:
                fw_name = model.lpad.get_fw_by_id(fw_id).name
                var_names.extend((v.name for v in vars_ if v.__fw_name == fw_name))
            msg += ', '.join(var_names)
            err_type = 'Ancestor error'
            logging.error('variable_value:%s ancestor error', repr(self))
            raise TextXError(msg, **get_location(self), err_type=err_type)
        raise NonCompletedException
    elif firework.state in ['READY', 'RESERVED', 'RUNNING']:
        raise NonCompletedException
    else:
        logging.critical('variable_value:%s unknown node state', repr(self))
        raise RuntimeError('Unknown node state')
    logging.debug('variable_value:%s', repr(self))
    return ret_value


def print_value(self):
    """evaluate the print function"""
    values = []
    for par in self.params:
        try:
            val = par.value
        except TextXError as err:
            if isinstance(err.__cause__, NonCompletedException):
                val = 'n.c.'
            else:
                raise err
        values.append(formatter(val))
    return ' '.join(values)


@textxerror_wrap
@checktype
def func_value(self):
    """evaluate a python function object"""
    logging.debug('func_value:%s', repr(self))
    func, pars = self.func
    assert all(isinstance_m(p, ['Variable', 'ObjectImport']) for p in pars)
    try:
        vals = [p.value for p in pars]
    except NonCompletedException:
        ret = 'n.c.'
    else:
        ret = get_basetyped(func(*vals))
    return ret


def get_lpad(self):
    """return launchpad object associated with the model"""
    return self._tx_model_params['model_instance']['lpad']


def get_uuid(self):
    """return workflow uuid associated with the model"""
    input_uuid = self._tx_model_params['model_instance']['uuid']
    return str(uuid.uuid4()) if input_uuid is None else input_uuid


def get_fw_ids(self):
    """updates and returns the list of nodes"""
    return self.lpad.get_fw_ids_in_wfs({'metadata.uuid': self.uuid})


def get_fw_ids_torun(self):
    """returns the list of nodes for which evaluation has been requested"""
    vars_ = set()
    for prnt in get_children_of_type('Print', self):
        vrefs = get_children_of_type('VarReference', prnt)
        vars_.update(vref.ref for vref in vrefs)
    fw_ids = []
    for var in iter(vars_):
        fw_ids.extend(self.lpad.get_fw_ids({'name': var.__fw_name}))
    anc_ids = [a for i in fw_ids for a in get_ancestors(self.lpad, i)]
    fw_ids_torun = list(set(fw_ids+anc_ids))
    assert all(f in self.fw_ids for f in fw_ids_torun)  # costs some overhead
    return fw_ids_torun


def add_workflow_properties(metamodel):
    """Add class properties using monkey style patching"""

    metamodel['Program'].value = property(program_value)
    metamodel['Program'].fw_ids = property(get_fw_ids)
    metamodel['Program'].fw_ids_torun = property(get_fw_ids_torun)
    metamodel['Program'].lpad = property(get_lpad)
    metamodel['Program'].uuid = cached_property(get_uuid)
    metamodel['Program'].uuid.__set_name__(metamodel['Program'], 'uuid')
    metamodel['Program'].workflow = cached_property(get_wf)
    metamodel['Program'].workflow.__set_name__(metamodel['Program'], 'workflow')

    metamodel['Print'].value = property(print_value)

    metamodel['Variable'].fireworks = cached_property(get_fws)
    metamodel['Variable'].fireworks.__set_name__(metamodel['Variable'], 'firework')
    metamodel['Variable']._get_fws_parallel = _get_fws_parallel
    metamodel['ObjectTo'].firework = cached_property(get_fw_object_to)
    metamodel['ObjectTo'].firework.__set_name__(metamodel['ObjectTo'], 'firework')

    mapping_dict = {
        'Variable': variable_value,
        'VarReference': func_value,
        'FunctionCall': func_value,
        'IfFunction': func_value,
        'IfExpression': func_value,
        'Expression': func_value,
        'Or': func_value,
        'Comparison': func_value,
        'Series': func_value,
        'Table': func_value,
        'IterableProperty': func_value,
        'IterableQuery': func_value,
        'ObjectImport': func_value,
        'Number': textxerror_wrap(number_value),
        'String': textxerror_wrap(plain_type_value),
        'Bool': textxerror_wrap(plain_type_value)
    }
    for key, func in mapping_dict.items():
        metamodel[key].value = cached_property(func)
        metamodel[key].value.__set_name__(metamodel[key], 'value')


def append_nodes(model, nodes):
    """recursively resolve dependencies in nodes and then append them"""
    while nodes:
        node = nodes.pop()
        parents = get_parent_nodes(model.lpad, model.fw_ids, node)
        if any(p is None for p in parents):
            append_nodes(model, nodes)
            parents = get_parent_nodes(model.lpad, model.fw_ids, node)
            assert all(p is not None for p in parents)
        model.lpad.append_wf(Workflow([node]), fw_ids=parents)


def workflow_model_processor(model, _):
    """generates a workflow for the just created model"""
    if not isinstance(model, str):
        model_params = getattr(model, '_tx_model_params')
        if model_params['model_instance']['uuid'] is None:
            logging.info('creating model from scratch')
            model.lpad.add_wf(model.workflow)
            logging.info('created model with UUID %s', model.uuid)
        else:
            logging.info('extending model with UUID %s', model.uuid)
            var_list = get_children_of_type('Variable', model)
            new_nodes = []
            for var in var_list:
                fw_ids = get_nodes_providing(model.lpad, model.fw_ids, var.name)
                if len(fw_ids) != 0:
                    assert len(fw_ids) == 1
                    var.__fw_name = model.lpad.get_fw_by_id(fw_ids[0]).name
                else:
                    new_nodes.extend(var.fireworks)
            append_nodes(model, new_nodes)
        if model_params.get('autorun'):
            logging.info('running model with UUID %s', model.uuid)

            def get_fws_torun():
                """return a list of fireworks if there are READY fireworks"""
                on_demand = model_params.get('on_demand')
                ids = model.fw_ids_torun if on_demand else model.fw_ids
                fw_query = {'fw_id': {'$in': ids}, 'state': 'READY'}
                if model.lpad.get_fw_ids(fw_query):
                    return ids
                return []
            fws_to_run = get_fws_torun()
            while fws_to_run:
                run_fireworks(model.lpad, fws_to_run)
                fws_to_run = get_fws_torun()
    else:
        logging.info('empty model')


def add_workflow_model_processor(metamodel):
    """register the remote model processor on the metamodel instance"""
    metamodel.register_model_processor(workflow_model_processor)
