# pylint: disable = protected-access
"""Metamodel manipulations in case of deferred evaluation"""
import uuid
import logging
from textx import textx_isinstance
from textx import get_children_of_type, get_parent_of_type

LOGLEVEL = logging.INFO
logging.basicConfig(level=LOGLEVEL)


def deferred_vartuple_processor(model, metamodel):
    """
    Replace the parameter of VarTuple objects with a VarReference in the case of
    deferred evaluation. This is performed to avoid repeated evaluations of the
    right hand side (the parameter) for each VarTuple Variable object.

    Example: This is equivalent to substitute '(a, b) = func(3)' with
    'c = func(3); (a, b) = c' where 'func(x) = (x, 2*x)'

    TODO: it would be nicer to pack (a, b) in a tuple of outputs in tuple_func()
    """
    if not isinstance(model, str):
        model_params = getattr(model, '_tx_model_params')
        if 'deferred_mode' in model_params:
            if model_params['deferred_mode']:
                for vart in get_children_of_type('VarTuple', model):
                    if textx_isinstance(vart.parameter, metamodel['Tuple']):
                        for var, par in zip(vart.variables, vart.parameter.params):
                            var.parameter = par
                    elif not textx_isinstance(vart.parameter, metamodel['VarReference']):
                        var = metamodel['Variable']()
                        setattr(var, 'name', str(uuid.uuid4()))
                        setattr(var, 'parameter', vart.parameter)
                        setattr(var, 'resources', None)
                        setattr(var, 'parent', model)
                        model.statements.append(var)
                        var_ref = metamodel['VarReference']()
                        setattr(var_ref, 'ref', var)
                        setattr(var_ref, 'parent', vart)
                        setattr(vart, 'parameter', var_ref)


def get_object_str(src, obj):
    """extract the string a certain textx model object from the model string"""
    return src[obj._tx_position:obj._tx_position_end].rstrip()


def source_code_variable_processor(model, metamodel):
    """add source code lines pertinent to the variable objects in the model"""
    if not isinstance(model, str):
        p_src = getattr(model, '_tx_model_params').get('source_code')
        for obj in get_children_of_type('Variable', model):
            if p_src:
                o_src = set()
                vartuple_obj = get_parent_of_type('VarTuple', obj)
                var_obj = vartuple_obj if vartuple_obj else obj
                o_src.add(get_object_str(p_src, var_obj))
                for vref in get_children_of_type('VarReference', obj):
                    if textx_isinstance(vref.ref, metamodel['ObjectImport']):
                        imps = get_parent_of_type('ObjectImports', vref.ref)
                        o_src.add(get_object_str(p_src, imps))
                for call in get_children_of_type('FunctionCall', obj):
                    func = call.function
                    if not textx_isinstance(func, metamodel['ObjectImport']):
                        o_src.add(get_object_str(p_src, func))
                setattr(obj, 'source_code', list(o_src))
            else:
                setattr(obj, 'source_code', None)
