"""
Register object and model processors

Model processors are callables that are called at the end of the parsing when
the whole model is instantiated. These processors accept the model and metamodel
as parameters.

Object processors are registered for particular classes (grammar rules) and are
called when objects of the given class are instantiated. Object processors
are registered by defining a map between a rule name and the callable that will
process the instances of the metamodel class correspoding to the rule. The rule
must be a match rule and no abstract rule.

Register object processors *only here*
"""
from virtmat.language.constraints.processors import add_constraints_processors
from virtmat.language.interpreter.instant_executor import add_io_processor
from virtmat.language.interpreter.workflow_executor import add_workflow_model_processor
from .function import function_call_processor
from .deferred import deferred_vartuple_processor
from .deferred import source_code_variable_processor
from .workflow import resources_processor
from .workflow import parallelizable_processor


def object_import_processor(obj):
    """add namespace of parent ObjectImports to ObjectImport"""
    obj.namespace = obj.parent.namespace


def table_processor(obj):
    """fill in the list of table columns"""
    if obj.columns_tuple:
        assert len(obj.columns) == 0
        obj.columns = obj.columns_tuple.params


def chem_term_processor(obj):
    """replace undefined coefficients with unity by convention"""
    if obj.coefficient is None:
        obj.coefficient = 1.0


def add_processors(metamodel, constr_processors=True, wflow_processors=False):
    """register the processors on the metamodel instance"""
    obj_processors = {'ObjectImport': object_import_processor,
                      'Resources': resources_processor,
                      'Map': parallelizable_processor,
                      'Filter': parallelizable_processor,
                      'Reduce': parallelizable_processor,
                      'Table': table_processor,
                      'ChemTerm': chem_term_processor}
    metamodel.register_obj_processors(obj_processors)
    metamodel.register_model_processor(source_code_variable_processor)
    metamodel.register_model_processor(function_call_processor)
    metamodel.register_model_processor(deferred_vartuple_processor)
    if constr_processors:
        add_constraints_processors(metamodel)
    if wflow_processors:
        add_workflow_model_processor(metamodel)
    else:
        add_io_processor(metamodel)
