"""handling domain specific errors"""
import sys
from textx import get_location, get_model
from textx.exceptions import TextXError, TextXSemanticError, TextXSyntaxError
from pint.errors import PintError, DimensionalityError, UndefinedUnitError
from ase.calculators.calculator import CalculatorSetupError


class InvalidUnitError(ValueError):
    """raise this exception if an invalid unit is detected"""


class ConvergenceError(RuntimeError):
    """raise this exception if a calculation has not converged"""


class StructureInputError(RuntimeError):
    """raise this exception if any exceptions are raised by ase.io.read"""


def print_stderr(*args, **kwargs):
    """print to stderr"""
    print(*args, file=sys.stderr, **kwargs)


def textxerror_wrap(func):
    """
    This is a decorator similar to the 'textxerror_wrap' from textX but it also
    sets the error context and accepts arbitrary list of arguments. The first
    positional argument must be a textX model object.
    """
    def wrapper(*args, **kwargs):
        obj = args[0]
        try:
            return func(*args, **kwargs)
        except Exception as err:
            if isinstance(err, TextXError):
                raise
            loc = get_location(obj)
            src = getattr(get_model(obj), '_tx_model_params').get('source_code')
            con = None if src is None else src[obj._tx_position:obj._tx_position_end].rstrip()
            raise TextXError(str(err), **loc, context=con) from err
    return wrapper


def process_error(err):
    """generic error processor for errors of class TextXError"""
    location = str(err.filename) + ':' + str(err.line) + ':' + str(err.col)
    if err.err_type is None:
        if isinstance(err, TextXSyntaxError):
            err_type = 'Syntax error'
        elif isinstance(err, TextXSemanticError):
            err_type = 'Semantic error'
        elif isinstance(err.__cause__, DimensionalityError):
            err_type = 'Dimensionality error'
        elif isinstance(err.__cause__, UndefinedUnitError):
            err_type = 'Undefined unit'
        elif isinstance(err.__cause__, PintError):
            err_type = 'Unknown units error'
        elif isinstance(err.__cause__, InvalidUnitError):
            err_type = 'Invalid units error'
        elif isinstance(err.__cause__, CalculatorSetupError):
            err_type = 'Calculator setup error'
        elif isinstance(err.__cause__, TypeError):
            err_type = 'Type error'
        elif isinstance(err.__cause__, IndexError):
            err_type = 'Index error'
        elif isinstance(err.__cause__, KeyError):
            err_type = 'Key error'
        elif isinstance(err.__cause__, ValueError):
            err_type = 'Value error'
        elif isinstance(err.__cause__, ArithmeticError):
            err_type = 'Arithmetic error'
        elif isinstance(err.__cause__, FileExistsError):
            err_type = 'File exists error'
        elif isinstance(err.__cause__, OSError):
            err_type = 'Operating system error'
        else:
            raise err.__cause__
    else:
        err_type = err.err_type

    msg = err_type + ': ' + location
    if err.context:
        msg += ' --> ' + err.context + ' <--'
    msg += '\n' + err.message
    print_stderr(msg)


def error_handler(func):
    """error handler decorator function"""
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except TextXError as err:
            process_error(err)
            return None
        except Exception as err:
            raise RuntimeError('non-handled exception') from err
    return wrapper
