"""serialization/deserialization code"""
import numpy
import pandas
import pint_pandas
from fireworks.utilities.fw_serializers import FWSerializable
from fireworks.utilities.fw_serializers import serialize_fw
from fireworks.utilities.fw_serializers import recursive_serialize
from fireworks.utilities.fw_serializers import recursive_deserialize
from virtmat.language.utilities import amml, chemistry
from .types import ureg
from .lists import list_apply, list_flatten


def convert_val(number):
    """convert to native python numerical types from numpy scalar types"""
    return getattr(number, 'item', lambda: number)()


def silence_dtype_warning_empty_series(func):
    """
    Silence this warning: FutureWarning: The default dtype for empty Series
    will be 'object' instead of 'float64' in a future version. Specify a dtype
    explicitly to silence this warning.
    To be applied to the __init__(self, *args, **kwargs) function
    """
    def dec_func(*args, **kwargs):
        if len(args) > 1:
            if len(args[1]) == 0:
                kwargs['dtype'] = 'object'
        elif len(kwargs['data']) == 0:
            kwargs['dtype'] = 'object'
        return func(*args, **kwargs)
    return dec_func


def series2dict(series):
    """convert pandas.Series with pint.Quantity to a dictionary"""
    if len(series) == 0:
        return {'data': [], 'name': series.name, 'units': None}
    units = [str(x.to_reduced_units().units) for x in series]
    unit = max(((u, units.count(u)) for u in set(units)), key=lambda c: c[1])[0]
    data = [convert_val(x.to(unit).magnitude) for x in series]
    return {'data': data, 'name': series.name, 'units': unit}


class FWDataFrame(pandas.DataFrame, FWSerializable):
    """serializable dataframe, i.e. with to_dict() and from_dict() methods"""
    _fw_name = '{{' + __loader__.name + '.' + __qualname__ + '}}'

    def __init__(self, *args, _fw_name=None, **kwargs):
        pandas.DataFrame.__init__(self, *args, **kwargs)

    @serialize_fw
    @recursive_serialize
    def to_dict(self):
        return {'data': [get_serializable(self[c]) for c in self.columns]}

    @classmethod
    @recursive_deserialize
    def from_dict(cls, m_dict):
        assert cls._fw_name == m_dict.pop('_fw_name')
        if len(m_dict['data']) == 0:
            return cls()
        return cls(pandas.concat(m_dict['data'], axis=1))

    def to_base(self):
        """return an instance of the base class"""
        return pandas.DataFrame(self)


class FWSeries(pandas.Series, FWSerializable):
    """serializable series, i.e. with to_dict() and from_dict() methods"""
    _fw_name = '{{' + __loader__.name + '.' + __qualname__ + '}}'

    @silence_dtype_warning_empty_series
    def __init__(self, *args, _fw_name=None, **kwargs):
        pandas.Series.__init__(self, *args, **kwargs)

    @serialize_fw
    @recursive_serialize
    def to_dict(self):
        dct = {'_fw_name': self._fw_name}
        if (isinstance(self.dtype, pint_pandas.PintType)
                or all(isinstance(e, ureg.Quantity) for e in self)):
            dct.update(series2dict(self))
        else:
            dct['name'] = self.name
            dct['data'] = [get_serializable(e) for e in self]
        return dct

    @classmethod
    @recursive_deserialize
    def from_dict(cls, m_dict):
        types = [type(elem) for elem in m_dict['data'] if elem is not None]
        if len(set(types)) > 1:
            if not all(t in (int, float) for t in types):
                raise TypeError('elements in series must have the same type')
        if float in types:
            units = m_dict.get('units', 'dimensionless')
            data = m_dict['data']
            dtype = pint_pandas.PintType(units)
        elif int in types:
            units = m_dict.get('units', None)
            elems = (numpy.nan if e is None else e for e in m_dict['data'])
            data = [ureg.Quantity(e, units) for e in elems]
            dtype = 'object'
        else:
            data = m_dict['data']
            dtype = None
        return cls(data=data, name=m_dict['name'], dtype=dtype)

    def to_base(self):
        """return an instance of the base class"""
        return pandas.Series(self)


class FWQuantity(FWSerializable, ureg.Quantity):
    """serializable pint quantity, i.e. with to_dict() and from_dict()"""
    _fw_name = '{{' + __loader__.name + '.' + __qualname__ + '}}'

    def __reduce__(self):
        return self.__class__, (self.magnitude, self.units)

    @serialize_fw
    @recursive_serialize
    def to_dict(self):
        dct = {'_fw_name': self._fw_name, 'data': self.to_tuple()}
        return dct

    @classmethod
    def from_dict(cls, m_dict):
        assert m_dict['_fw_name'] == cls._fw_name
        assert isinstance(m_dict['data'], (list, tuple))
        return super().from_tuple(m_dict['data'])

    def to_base(self):
        """return an instance of the base class"""
        return ureg.Quantity.from_tuple((self.to_tuple()))


class FWBoolArray(numpy.ndarray, FWSerializable):
    """JSON serializable bool numpy.ndarray, i.e. with to_dict() and from_dict()"""
    _fw_name = '{{' + __loader__.name + '.' + __qualname__ + '}}'

    def __new__(cls, data):
        return numpy.asarray(data).view(cls)

    @serialize_fw
    @recursive_serialize
    def to_dict(self):
        data = self.tolist()
        return {'_fw_name': self._fw_name, 'data': data}

    @classmethod
    def from_dict(cls, m_dict):
        assert m_dict['_fw_name'] == cls._fw_name
        assert isinstance(m_dict['data'], list)
        assert all(isinstance(e, bool) for e in list_flatten(m_dict['data']))
        return cls(m_dict['data'])

    def to_base(self):
        """return an instance of the base class"""
        return numpy.array(self)


class FWStrArray(numpy.ndarray, FWSerializable):
    """JSON serializable str numpy.ndarray, i.e. with to_dict() and from_dict()"""
    _fw_name = '{{' + __loader__.name + '.' + __qualname__ + '}}'

    def __new__(cls, data):
        return numpy.asarray(data).view(cls)

    @serialize_fw
    @recursive_serialize
    def to_dict(self):
        data = self.tolist()
        return {'_fw_name': self._fw_name, 'data': data}

    @classmethod
    def from_dict(cls, m_dict):
        assert m_dict['_fw_name'] == cls._fw_name
        assert isinstance(m_dict['data'], list)
        assert all(isinstance(e, str) for e in list_flatten(m_dict['data']))
        return cls(m_dict['data'])

    def to_base(self):
        """return an instance of the base class"""
        return numpy.array(self)


class FWIntArray(numpy.ndarray, FWSerializable):
    """JSON serializable int numpy.ndarray, i.e. with to_dict() and from_dict()"""
    _fw_name = '{{' + __loader__.name + '.' + __qualname__ + '}}'

    def __new__(cls, data):
        data_check = list(data.flat) if isinstance(data, numpy.ndarray) else list_flatten(data)
        assert all(isinstance(e, ureg.Quantity) for e in data_check)
        assert len(set(str(e.units) for e in data_check)) == 1
        return numpy.asarray(data, dtype='object').view(cls)

    @serialize_fw
    @recursive_serialize
    def to_dict(self):
        units = str(next(e.units for e in self.flat))
        data = list_apply(lambda x: x.magnitude, self.tolist())
        return {'_fw_name': self._fw_name, 'data': data, 'units': units}

    @classmethod
    def from_dict(cls, m_dict):
        assert m_dict['_fw_name'] == cls._fw_name
        assert isinstance(m_dict['data'], list)
        units = m_dict.get('units')
        data = list_apply(lambda x: ureg.Quantity(x, units), m_dict['data'])
        return cls(data)

    def to_base(self):
        """return an instance of the base class"""
        return numpy.array(self)


class FWFloatArray(pint_pandas.PintArray, FWSerializable):
    """JSON serializable PintArray, i.e. with to_dict() and from_dict()"""
    _fw_name = '{{' + __loader__.name + '.' + __qualname__ + '}}'

    @serialize_fw
    @recursive_serialize
    def to_dict(self):
        dct = {'_fw_name': self._fw_name, 'units': str(self.units),
               'data': self.data.tolist()}
        return dct

    @classmethod
    def from_dict(cls, m_dict):
        assert m_dict['_fw_name'] == cls._fw_name
        units = m_dict.get('units')
        assert isinstance(units, str)
        assert isinstance(m_dict['data'], list)
        return cls(m_dict['data'], dtype=pint_pandas.PintType(units))

    def to_base(self):
        """return an instance of the base class"""
        return pint_pandas.PintArray(self.data, dtype=self.dtype)

    @classmethod
    def from_base(cls, obj):
        """return an instance from an instance of the base class"""
        return cls(obj.data, dtype=obj.dtype)


class FWAMMLStructure(amml.AMMLStructure, FWSerializable):
    """JSON serializable amml.AMMLStructure, i.e. with to_dict() and from_dict()"""
    _fw_name = '{{' + __loader__.name + '.' + __qualname__ + '}}'

    @serialize_fw
    @recursive_serialize
    def to_dict(self):
        return {'_fw_name': self._fw_name, 'data': FWDataFrame(self.tab),
                'name': self.name}

    @classmethod
    @recursive_deserialize
    def from_dict(cls, m_dict):
        assert m_dict['_fw_name'] == cls._fw_name
        return cls(m_dict['data'], m_dict['name'])

    def to_base(self):
        """return an instance of the base class"""
        return amml.AMMLStructure(get_basetyped(self.tab), self.name)

    @classmethod
    def from_base(cls, obj):
        """return an instance from an instance of the base class"""
        return cls(obj.tab, obj.name)


class FWCalculator(amml.Calculator, FWSerializable):
    """JSON serializable amml.Calculator, i.e. with to_dict() and from_dict()"""
    _fw_name = '{{' + __loader__.name + '.' + __qualname__ + '}}'

    @serialize_fw
    @recursive_serialize
    def to_dict(self):
        return {'_fw_name': self._fw_name, 'parameters': get_serializable(self.parameters),
                'name': self.name, 'pinning': self.pinning, 'version': self.version}

    @classmethod
    @recursive_deserialize
    def from_dict(cls, m_dict):
        assert cls._fw_name == m_dict.pop('_fw_name')
        return cls(**m_dict)

    def to_base(self):
        """return an instance of the base class"""
        return amml.Calculator(self.name, get_basetyped(self.parameters),
                               pinning=self.pinning, version=self.version)

    @classmethod
    def from_base(cls, obj):
        """return an instance from an instance of the base class"""
        return cls(obj.name, obj.parameters, pinning=obj.pinning, version=obj.version)


class FWProperty(amml.Property, FWSerializable):
    """JSON serializable amml.Property, i.e. with to_dict() and from_dict()"""
    _fw_name = '{{' + __loader__.name + '.' + __qualname__ + '}}'

    @serialize_fw
    @recursive_serialize
    def to_dict(self):
        return {'_fw_name': self._fw_name,
                'names': get_serializable(self.names),
                'calc': get_serializable(self.calc),
                'struct': get_serializable(self.struct),
                'constrs': get_serializable(self.constrs),
                'task': get_serializable(self.task),
                'results': get_serializable(self.results)}

    @classmethod
    @recursive_deserialize
    def from_dict(cls, m_dict):
        assert cls._fw_name == m_dict.pop('_fw_name')
        return cls(**m_dict)

    def to_base(self):
        """return an instance of the base class"""
        return amml.Property(get_basetyped(self.names), get_basetyped(self.calc),
                             get_basetyped(self.struct),
                             constrs=get_basetyped(self.constrs),
                             task=get_basetyped(self.task),
                             results=get_basetyped(self.results))

    @classmethod
    def from_base(cls, obj):
        """return an instance from an instance of the base class"""
        return cls(obj.names, obj.calc, obj.struct, constrs=obj.constrs,
                   task=obj.task, results=obj.results)


class FWConstraint(amml.Constraint, FWSerializable):
    """JSON serializable amml.Constraint, i.e. with to_dict() and from_dict()"""
    _fw_name = '{{' + __loader__.name + '.' + __qualname__ + '}}'

    @serialize_fw
    @recursive_serialize
    def to_dict(self):
        ser_kwargs = {k: get_serializable(v) for k, v in self.kwargs.items()}
        return {'_fw_name': self._fw_name, 'name': self.name, **ser_kwargs}

    @classmethod
    @recursive_deserialize
    def from_dict(cls, m_dict):
        assert cls._fw_name == m_dict.pop('_fw_name')
        return cls(**m_dict)

    def to_base(self):
        """return an instance of the base class"""
        return amml.Constraint(self.name, **self.kwargs)

    @classmethod
    def from_base(cls, obj):
        """return an instance from an instance of the base class"""
        return cls(obj.name, **obj.kwargs)


class FWChemSpecies(chemistry.ChemSpecies, FWSerializable):
    """JSON serializable chemistry.ChemSpecies, i.e. with to_dict() and from_dict()"""
    _fw_name = '{{' + __loader__.name + '.' + __qualname__ + '}}'

    @serialize_fw
    @recursive_serialize
    def to_dict(self):
        return {'_fw_name': self._fw_name,
                'name': self.name,
                'props': get_serializable(self.props)}

    @classmethod
    @recursive_deserialize
    def from_dict(cls, m_dict):
        assert cls._fw_name == m_dict.pop('_fw_name')
        return cls(**m_dict)

    def to_base(self):
        """return an instance of the base class"""
        return chemistry.ChemSpecies(self.name, self.props)

    @classmethod
    def from_base(cls, obj):
        """return an instance from an instance of the base class"""
        return cls(obj.name, obj.props)


class FWChemReaction(chemistry.ChemReaction, FWSerializable):
    """JSON serializable chemistry.ChemReaction, i.e. with to_dict() and from_dict()"""
    _fw_name = '{{' + __loader__.name + '.' + __qualname__ + '}}'

    @serialize_fw
    @recursive_serialize
    def to_dict(self):
        return {'_fw_name': self._fw_name,
                'terms': get_serializable(self.terms),
                'props': get_serializable(self.props)}

    @classmethod
    @recursive_deserialize
    def from_dict(cls, m_dict):
        assert cls._fw_name == m_dict.pop('_fw_name')
        return cls(**m_dict)

    def to_base(self):
        """return an instance of the base class"""
        return chemistry.ChemReaction(self.terms, self.props)

    @classmethod
    def from_base(cls, obj):
        """return an instance from an instance of the base class"""
        return cls(obj.terms, obj.props)


def get_serializable(obj):
    """convert an arbitrary Python object to a JSON serializable object"""
    if not isinstance(obj, FWSerializable):
        if isinstance(obj, ureg.Quantity):
            retval = FWQuantity.from_tuple(obj.to_tuple())
        elif isinstance(obj, pandas.DataFrame):
            retval = FWDataFrame(obj)
        elif isinstance(obj, pandas.Series):
            retval = FWSeries(obj)
        elif isinstance(obj, pint_pandas.PintArray):
            retval = FWFloatArray.from_base(obj)
        elif isinstance(obj, numpy.ndarray):
            if obj.dtype.type is numpy.bool_:
                retval = FWBoolArray(obj)
            elif obj.dtype.type is numpy.str_:
                retval = FWStrArray(obj)
            else:
                assert obj.dtype == numpy.dtype('object')
                retval = FWIntArray(obj)
        elif isinstance(obj, amml.AMMLStructure):
            retval = FWAMMLStructure.from_base(obj)
        elif isinstance(obj, amml.Calculator):
            retval = FWCalculator.from_base(obj)
        elif isinstance(obj, amml.Property):
            retval = FWProperty.from_base(obj)
        elif isinstance(obj, amml.Constraint):
            retval = FWConstraint.from_base(obj)
        elif isinstance(obj, chemistry.ChemSpecies):
            retval = FWChemSpecies.from_base(obj)
        elif isinstance(obj, chemistry.ChemReaction):
            retval = FWChemReaction.from_base(obj)
        elif isinstance(obj, (list, tuple)):
            retval = [get_serializable(o) for o in obj]
        elif isinstance(obj, dict):
            retval = {k: get_serializable(v) for k, v in obj.items()}
        else:
            json_types = (bool, int, float, str, type(None))
            if not isinstance(obj, json_types):
                msg = f'cannot serialize {obj} with type {type(obj)}'
                raise ValueError(msg)
            retval = obj
    else:
        retval = obj
    return retval


def get_basetyped(obj):
    """converts an JSON serializable object to an object of its base class"""
    if isinstance(obj, FWSerializable):
        retval = obj.to_base()
    elif isinstance(obj, (list, tuple)):
        retval = tuple(get_basetyped(o) for o in obj)
    else:
        retval = obj
    return retval
