"""Utility functions for textx"""
import re
from os import path
from textx import get_metamodel
from textx import textx_isinstance


def isinstance_m(obj, classes):
    """
    Check whether an object is an instance of metamodel classes
    Args:
        obj: a model object
        classes: an iterable with names of classes from the metamodel
    Returns:
        True (False) if the object is (is not) instance of any class
    """
    meta = get_metamodel(obj)
    return any(textx_isinstance(obj, meta[c]) for c in classes)


def isinstance_r(obj, classes):
    """
    Check whether an instance of a class from classes is referenced in obj
    Args:
        obj: a model object
        classes: an iterable with names of classes from the metamodel
    Returns:
        True if an instance of the class is referenced, otherwise False
    """
    if hasattr(obj, 'ref'):
        ret = isinstance_r(obj.ref, classes)
    elif isinstance_m(obj, ['Variable']):
        ret = isinstance_r(obj.parameter, classes)
    else:
        ret = isinstance_m(obj, classes)
    return ret


def is_reference(obj, metamodel):
    """return True if obj is a VarReference"""
    return textx_isinstance(obj, metamodel['VarReference'])


def get_reference(obj):
    """return the referenced object if obj is a reference and obj otherwise"""
    return obj.ref.parameter if is_reference(obj, get_metamodel(obj)) else obj


class GrammarString:
    """create a single textX grammar string from a modular textX grammar"""
    __regex = r'^import\s+(\S+)$'

    def __init__(self, grammar_path):
        self.grammar_dir = path.dirname(grammar_path)
        self.__memo = set()
        self.__string = ''.join(self._expand_grammar(grammar_path))

    @property
    def string(self):
        """string getter"""
        return self.__string

    def _expand_grammar(self, filename):
        """recursively expand all imported grammar files without duplicates"""
        with open(filename, 'r', encoding='utf-8') as inp:
            lines = inp.readlines()
        new_lines = []
        inc_lines = []
        for line in lines:
            match = re.search(self.__regex, line)
            if match:
                include = match.group(1).replace('.', '/') + '.tx'
                if include not in self.__memo:
                    self.__memo.add(include)
                    include_file = path.join(self.grammar_dir, include)
                    inc_lines.extend(self._expand_grammar(include_file))
            else:
                new_lines.append(line)
        new_lines.extend(inc_lines)
        return new_lines
