"""test the session class"""
import os
import time
import pytest
from fireworks import LaunchPad
from fireworks.fw_config import LAUNCHPAD_LOC
from textx import get_children_of_type
from textx.exceptions import TextXSemanticError, TextXError
from virtmat.language.interpreter.session_manager import Session
from virtmat.language.utilities.errors import error_handler
from virtmat.language.interpreter.workflow_executor import NonCompletedException

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
GRAMMAR_PATH = os.path.join(ROOT_DIR, '..', 'grammar', 'virtmat.tx')


@pytest.fixture(name='lpad')
def lpad_fixture():
    """launchpad object as fixture for all tests"""
    return LaunchPad.from_file(LAUNCHPAD_LOC) if LAUNCHPAD_LOC else LaunchPad()


def test_session_model_from_scratch_from_empty_model_str(lpad):
    """start a new session with a new model and empty model string"""
    session = Session(lpad, uuid=None, grammar_path=GRAMMAR_PATH,
                      model_str='', autorun=True)
    assert session.model is None
    assert session.get_model() is None
    assert session.get_model_str() == ''
    assert session.uuid is None


def test_session_model_from_scratch_from_model_str_comments_blanks(lpad):
    """start a new session with a new model with only comments and blanks"""
    session = Session(lpad, uuid=None, grammar_path=GRAMMAR_PATH,
                      model_str=' ', autorun=True)
    assert session.model is None
    assert session.get_model() is None
    assert session.get_model_str() == ''
    assert session.uuid is None
    assert session.get_model('# comment; """long comment"""') is None
    assert session.get_model_str() == ''
    assert session.uuid is None


def test_session_model_from_scratch_from_empty_model_str_get_model_empty(lpad):
    """start a new session with a new model and empty model string and call
       get_model() with an empty string"""
    session = Session(lpad, uuid=None, grammar_path=GRAMMAR_PATH,
                      model_str='', autorun=True)
    assert session.model is None
    assert session.get_model('') is None
    assert session.get_model_str() == ''
    assert session.uuid is None


def test_session_model_from_scratch_from_none_model_str_get_model_none(lpad):
    """start a new session with a new model and None model string and call
       get_model() with None model string"""
    session = Session(lpad, uuid=None, grammar_path=GRAMMAR_PATH, autorun=True)
    assert session.model is None
    assert session.get_model() is None
    assert session.get_model_str() == ''
    assert session.uuid is None


def test_session_model_from_scratch_from_none_model_str(lpad):
    """start a new session with a new model and None model string and call
       get_model() with None model string"""
    session = Session(lpad, uuid=None, grammar_path=GRAMMAR_PATH, autorun=True)
    assert session.model is None
    assert session.get_model('a = 1; print(a)').value == '1'
    assert session.get_model_str() == 'a = 1\n'
    assert session.uuid is not None
    assert session.get_model() is session.get_model()  # model not reinstantiated


def test_session_model_from_scratch_from_empty_model_extend(lpad):
    """start a new session with a new model and empty model string; extend the
       model twice in the same session"""
    session = Session(lpad, uuid=None, grammar_path=GRAMMAR_PATH,
                      model_str='', autorun=True)
    assert session.model is None
    assert session.get_model() is None
    assert session.get_model('a = 1').value == ''
    assert session.get_model('print(a)').value == '1'
    assert session.get_model_str() == 'a = 1\n'
    assert session.uuid is not None


def test_session_model_from_scratch_from_model_str_novar(lpad):
    """a session with a new model without variables; then extend the model in
       the same session with a model without variables"""
    session = Session(lpad, uuid=None, grammar_path=GRAMMAR_PATH,
                      model_str='print(1)', autorun=True)
    assert session.model is not None
    assert session.get_model().value == '1'
    assert session.get_model('print(2)').value == '2'
    assert session.uuid is not None


def test_session_model_from_scratch_from_model_str_no_autorun(lpad):
    """a session with a new model with a variable; then extend the model with a
       print statement with autorun deactivated"""
    session = Session(lpad, uuid=None, grammar_path=GRAMMAR_PATH,
                      model_str='a = 1', autorun=False)
    assert session.model is not None
    assert session.get_model().value == ''
    assert session.get_model('print(a)').value == 'n.c.'
    assert session.uuid is not None


def test_session_model_from_scratch_from_model_str_autorun(lpad):
    """"a session with a new model; extend the model twice with autorun"""
    session = Session(lpad, uuid=None, grammar_path=GRAMMAR_PATH,
                      model_str='a = 1', autorun=True)
    assert session.get_model().value == ''
    assert session.get_model('print(a)').value == '1'
    assert session.get_model('b = a + 2; print(b)').value == '3'
    assert session.uuid is not None


def test_session_model_from_scratch_from_model_str_single_import(lpad):
    """a session with a new model with just an import; then extend with an
       import reference showing that a model cannot be extended with just an
       import"""
    session = Session(lpad, uuid=None, grammar_path=GRAMMAR_PATH,
                      model_str='use math.pi', autorun=True)
    msg = 'Unknown object "pi" of class "GeneralVariable"'
    with pytest.raises(TextXSemanticError, match=msg):
        session.get_model('print(pi)')


def test_session_model_from_scratch_from_model_str_single_function_def(lpad):
    """a session with a new model with just a function def; then extend with an
       function def reference (a function call) showing that a model cannot be
       extended with just an function definition"""
    session = Session(lpad, uuid=None, grammar_path=GRAMMAR_PATH,
                      model_str='f(x) = 2*x', autorun=True)
    msg = 'Unknown object "f" of class "Function"'
    with pytest.raises(TextXSemanticError, match=msg):
        session.get_model('print(f(1))')


def test_session_model_from_scratch_from_model_str_import(lpad):
    """a session with a new model with an import and two import references;
       then extend twice with further references"""
    session = Session(lpad, uuid=None, grammar_path=GRAMMAR_PATH,
                      model_str='use math.pi; print(pi); a = pi', autorun=True)
    pi_ref = pytest.approx(3.141592653589793)
    assert float(session.get_model().value) == pi_ref
    assert float(session.get_model('print(a)').value) == pi_ref
    assert float(session.get_model('print(pi)').value) == pi_ref


def test_session_model_from_scratch_from_model_str_function_def(lpad):
    """a session with a new model with a function def and two references; then
       extend twice with two references"""
    session = Session(lpad, uuid=None, grammar_path=GRAMMAR_PATH,
                      model_str='f(x) = 2*x; a = f(1)', autorun=True)
    assert session.get_model('print(f(2))').value == '4'
    assert session.get_model('print(a)').value == '2'


def test_session_model_from_scratch_from_model_str_vartuple(lpad):
    """a session with a new model with a vartuple; then extend with a
       reference"""
    session = Session(lpad, uuid=None, grammar_path=GRAMMAR_PATH,
                      model_str='(a, b) = (1, 2)', autorun=True)
    assert session.get_model('print(a, b)').value == '1 2'


def test_session_model_with_no_uuid_and_no_grammar(lpad):
    """a session with neither uuid nor grammar provided"""
    msg = 'grammar must be provided if uuid not provided'
    with pytest.raises(ValueError, match=msg):
        Session(lpad)


def test_session_model_with_invalid_uuid(lpad):
    """a session with an invalid uuid, i.e. no model exists with the uuid"""
    msg = 'no model with provided uuid could be found'
    with pytest.raises(RuntimeError, match=msg):
        Session(lpad, uuid='f28d2b56-68e9-4dd8-844b-726a54f54aa2')


def test_session_model_with_uuid_and_grammar(lpad):
    """a session with both uuid and grammar provided"""
    old_session = Session(lpad, grammar_path=GRAMMAR_PATH, model_str='a = 1')
    uuid = old_session.uuid
    assert uuid is not None
    msg = 'provided grammar ignored in favor of grammar from provided uuid'
    with pytest.warns(UserWarning, match=msg):
        Session(lpad, uuid=uuid, grammar_path=GRAMMAR_PATH)


def test_develop_a_model_in_two_sessions(lpad):
    """open a session with a model from scratch and then extend the same model
       in a second session"""
    m_str1 = 'func(x) = b*x; b = 3; a = func(2)'
    session1 = Session(lpad, grammar_path=GRAMMAR_PATH, model_str=m_str1)
    assert session1.model is not None
    assert session1.model.value == ''
    assert session1.get_model().value == ''
    session2 = Session(lpad, uuid=session1.uuid, autorun=True)
    assert session2.model is not None
    assert session2.model is session2.get_model()
    assert session2.get_model() is session2.get_model()
    assert session2.get_model().value == ''
    m_str2 = 'c = a + b + func(3); print(a, c)'
    assert session2.get_model(m_str2).value == '6 18'


def test_model_with_no_reference_in_extension_model(lpad):
    """extend a model with a model containing no reference"""
    session = Session(lpad, grammar_path=GRAMMAR_PATH, model_str='a = 1')
    session.get_model('b = 2')
    assert session.get_model('print(a, b)').value == 'n.c. n.c.'


def test_extend_with_a_model_with_random_order_of_references(lpad):
    """extend a model with a model with random order of references"""
    session = Session(lpad, grammar_path=GRAMMAR_PATH, model_str='a = 1', autorun=True)
    session.get_model('b = c; c = a')
    assert session.get_model('print(a, b, c)').value == '1 1 1'


def test_error_handler_static_errors(lpad, capsys):
    """test the domain specific error handler for static errors"""
    session = Session(lpad, grammar_path=GRAMMAR_PATH, autorun=False)
    error_handler(session.get_model)('a = b()')
    err_msg = 'Unknown object: None:1:5\nUnknown object "b" of class "Function"\n'
    std_err = capsys.readouterr().err
    assert std_err == err_msg
    error_handler(session.get_model)(';')
    err_msg = ("Syntax error: None:1:1 --> *; <--\nExpected '#.*$' or 'use' or "
               "'(' or ID or 'print' or EOF\n")
    std_err = capsys.readouterr().err
    assert std_err == err_msg


def test_error_handler_runtime_errors(lpad, capsys):
    """test the domain specific error handler for runtime errors"""
    session = Session(lpad, grammar_path=GRAMMAR_PATH, autorun=True)
    model = session.get_model('a = 4*(2*1 - 3/0); print(a)')
    capsys.readouterr()
    error_handler(lambda: model.value)()
    err_msg = 'Arithmetic error: None:1:1 --> a = 4*(2*1 - 3/0) <--\ndivision by zero\n'
    std_err = capsys.readouterr().err
    assert std_err == err_msg
    model = session.get_model('b = 1 [m] > 20 [sec]; print(b)')
    capsys.readouterr()
    error_handler(lambda: model.value)()
    err_msg = ("Dimensionality error: None:2:1 --> b = 1 [m] > 20 [sec] <--\n"
               "Cannot convert from 'meter' ([length]) to 'second' ([time])\n")
    std_err = capsys.readouterr().err
    assert std_err == err_msg


def test_error_handler_runtime_ancestor_error(lpad, capsys):
    """test the domain specific error handler for ancestor runtime errors"""
    session = Session(lpad, grammar_path=GRAMMAR_PATH, autorun=True)
    model = session.get_model('a = 1/0; b = a; print(b)')
    capsys.readouterr()
    error_handler(lambda: model.value)()
    err_msg = ('Ancestor error: None:1:10\nEvaluation not possible due to failed'
               ' ancestors: a\n')
    std_err = capsys.readouterr().err
    assert std_err == err_msg


def test_execution_on_demand(lpad):
    """test execution on demand"""
    session = Session(lpad, grammar_path=GRAMMAR_PATH, autorun=True,
                      on_demand=True, sleep_time=1)
    model = session.get_model('a = 1; b = a; c = 2; print(b)')
    var_objs = get_children_of_type('Variable', model)
    assert next(v.value for v in var_objs if v.name == 'a') == 1
    assert next(v.value for v in var_objs if v.name == 'b') == 1
    with pytest.raises(TextXError) as err_info:
        next(v.value for v in var_objs if v.name == 'c')
    assert isinstance(err_info.value.__cause__, NonCompletedException)
    model = session.get_model('print(a, b, c)')
    assert next(v.value for v in var_objs if v.name == 'a') == 1
    assert next(v.value for v in var_objs if v.name == 'b') == 1
    assert next(v.value for v in var_objs if v.name == 'c') == 2


def test_asynchronous_execution(lpad):
    """test asynchronous execution using WFEngine"""
    session = Session(lpad, grammar_path=GRAMMAR_PATH, autorun=True,
                      async_run=True, sleep_time=1)
    session.stop_runner()
    assert session.get_model('a = 1; print(a)').value == 'n.c.'
    session.start_runner()
    time.sleep(2)
    assert session.get_model('print(a)').value == '1'
    session.stop_runner()
    assert session.get_model('b = true').value == ''
    time.sleep(2)
    assert session.get_model('print(b)').value == 'n.c.'
    session.start_runner()
    time.sleep(2)
    assert session.get_model('print(b)').value == 'true'
    session.stop_runner()


def test_asynchronous_execution_on_demand(lpad):
    """test asynchronous execution on demand using WFEngine"""
    session = Session(lpad, grammar_path=GRAMMAR_PATH, autorun=True,
                      async_run=True, on_demand=True, sleep_time=1)
    session.stop_runner()
    assert session.get_model('a = 1; b = a; c = 2; print(b)').value == 'n.c.'
    session.start_runner()
    time.sleep(2)
    session.stop_runner()
    assert session.get_model('print(a, b, c)').value == '1 1 n.c.'
    session.start_runner()
    time.sleep(2)
    assert session.get_model('print(a, b, c)').value == '1 1 2'
    session.stop_runner()
